@@ -1,20 +1,8 @@
 /*
- *  Copyright (c) 2006 Boudewijn Rempt <boud@valdyas.org>
- *            (c) 2009 Dmitry Kazakov <dimula73@gmail.com>
+ *  SPDX-FileCopyrightText: 2006 Boudewijn Rempt <boud@valdyas.org>
+ *  SPDX-FileCopyrightText: 2009 Dmitry Kazakov <dimula73@gmail.com>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "kis_mask.h"
@@ -43,6 +31,8 @@
 #include "kis_mask_projection_plane.h"
 
 #include "kis_raster_keyframe_channel.h"
+#include "KisSafeNodeProjectionStore.h"
+
 
 struct Q_DECL_HIDDEN KisMask::Private {
     Private(KisMask *_q)
@@ -60,22 +50,24 @@ struct Q_DECL_HIDDEN KisMask::Private {
      * device belongs to the node, but not to the device itself.  So
      * the offset is set when the node is created, but not when the
      * selection is initialized. This causes the X,Y values to be
-     * lost, since the selection doen not exist at the moment. That is
+     * lost, since the selection does not exist at the moment. That is
      * why we save it separately.
      */
     QScopedPointer<QPoint> deferredSelectionOffset;
 
     KisAbstractProjectionPlaneSP projectionPlane;
-    KisCachedSelection cachedSelection;
+    KisSafeSelectionNodeProjectionStoreSP safeProjection;
 
-    void initSelectionImpl(KisSelectionSP copyFrom, KisLayerSP parentLayer, KisPaintDeviceSP copyFromDevice);
+    void initSelectionImpl(KisSelectionSP copyFrom, KisLayerSP parentLayer, KisPaintDeviceSP copyFromDevice, KisImageWSP image);
 };
 
-KisMask::KisMask(const QString & name)
-        : KisNode()
+KisMask::KisMask(KisImageWSP image, const QString &name)
+        : KisNode(image)
         , m_d(new Private(this))
 {
     setName(name);
+    m_d->safeProjection = new KisSafeSelectionNodeProjectionStore();
+    m_d->safeProjection->setImage(image);
 }
 
 KisMask::KisMask(const KisMask& rhs)
@@ -85,6 +77,8 @@ KisMask::KisMask(const KisMask& rhs)
 {
     setName(rhs.name());
 
+    m_d->safeProjection = new KisSafeSelectionNodeProjectionStore(*rhs.m_d->safeProjection);
+
     if (rhs.m_d->selection) {
         m_d->selection = new KisSelection(*rhs.m_d->selection.data());
         m_d->selection->setParentNode(this);
@@ -99,16 +93,37 @@ KisMask::KisMask(const KisMask& rhs)
 
 KisMask::~KisMask()
 {
+    if (m_d->selection) {
+        m_d->selection->setParentNode(0);
+    }
+
     delete m_d;
 }
 
 void KisMask::setImage(KisImageWSP image)
 {
     KisPaintDeviceSP parentPaintDevice = parent() ? parent()->original() : 0;
-    KisDefaultBoundsBaseSP defaultBounds = new KisSelectionDefaultBounds(parentPaintDevice, image);
+    KisDefaultBoundsBaseSP defaultBounds;
+
+    if (parentPaintDevice) {
+        defaultBounds = new KisSelectionDefaultBounds(parentPaintDevice);
+    } else {
+        if (image) {
+            qWarning() << "WARNING: KisMask::setImage() was called without any parent layer being set";
+        }
+
+        // just a fallback solution
+        defaultBounds = new KisDefaultBounds(image);
+    }
+
     if (m_d->selection) {
         m_d->selection->setDefaultBounds(defaultBounds);
+        m_d->selection->setResolutionProxy(m_d->selection->resolutionProxy()->createOrCloneDetached(image));
     }
+
+    m_d->safeProjection->setImage(image);
+
+    KisNode::setImage(image);
 }
 
 bool KisMask::allowAsChild(KisNodeSP node) const
@@ -141,20 +156,20 @@ const KoCompositeOp * KisMask::compositeOp() const
 
 void KisMask::initSelection(KisSelectionSP copyFrom, KisLayerSP parentLayer)
 {
-    m_d->initSelectionImpl(copyFrom, parentLayer, 0);
+    m_d->initSelectionImpl(copyFrom, parentLayer, 0, image());
 }
 
 void KisMask::initSelection(KisPaintDeviceSP copyFromDevice, KisLayerSP parentLayer)
 {
-    m_d->initSelectionImpl(0, parentLayer, copyFromDevice);
+    m_d->initSelectionImpl(0, parentLayer, copyFromDevice, image());
 }
 
 void KisMask::initSelection(KisLayerSP parentLayer)
 {
-    m_d->initSelectionImpl(0, parentLayer, 0);
+    m_d->initSelectionImpl(0, parentLayer, 0, image());
 }
 
-void KisMask::Private::initSelectionImpl(KisSelectionSP copyFrom, KisLayerSP parentLayer, KisPaintDeviceSP copyFromDevice)
+void KisMask::Private::initSelectionImpl(KisSelectionSP copyFrom, KisLayerSP parentLayer, KisPaintDeviceSP copyFromDevice, KisImageWSP image)
 {
     Q_ASSERT(parentLayer);
 
@@ -165,19 +180,28 @@ void KisMask::Private::initSelectionImpl(KisSelectionSP copyFrom, KisLayerSP par
          * We can't use setSelection as we may not have parent() yet
          */
         selection = new KisSelection(*copyFrom);
-        selection->setDefaultBounds(new KisSelectionDefaultBounds(parentPaintDevice, parentLayer->image()));
-        if (copyFrom->hasShapeSelection()) {
-            delete selection->flatten();
-        }
+        selection->setDefaultBounds(new KisSelectionDefaultBounds(parentPaintDevice));
+        selection->setResolutionProxy(copyFrom->resolutionProxy()->createOrCloneDetached(image));
     } else if (copyFromDevice) {
-        selection = new KisSelection(new KisSelectionDefaultBounds(parentPaintDevice, parentLayer->image()));
+        KritaUtils::DeviceCopyMode copyMode =
+            q->inherits("KisFilterMask") || q->inherits("KisTransparencyMask") ?
+            KritaUtils::CopyAllFrames : KritaUtils::CopySnapshot;
 
-        QRect rc(copyFromDevice->extent());
-        KisPainter::copyAreaOptimized(rc.topLeft(), copyFromDevice, selection->pixelSelection(), rc);
-        selection->pixelSelection()->invalidateOutlineCache();
+        selection = new KisSelection(copyFromDevice, copyMode,
+                                     new KisSelectionDefaultBounds(parentPaintDevice),
+                                     toQShared(new KisImageResolutionProxy(image)));
 
+        KisPixelSelectionSP pixelSelection = selection->pixelSelection();
+        if (pixelSelection->framesInterface()) {
+            KisRasterKeyframeChannel *keyframeChannel = pixelSelection->keyframeChannel();
+            keyframeChannel->setFilenameSuffix(".pixelselection");
+
+            q->addKeyframeChannel(keyframeChannel);
+            q->enableAnimation();
+        }
     } else {
-        selection = new KisSelection(new KisSelectionDefaultBounds(parentPaintDevice, parentLayer->image()));
+        selection = new KisSelection(new KisSelectionDefaultBounds(parentPaintDevice),
+                                     toQShared(new KisImageResolutionProxy(image)));
         selection->pixelSelection()->setDefaultPixel(KoColor(Qt::white, selection->pixelSelection()->colorSpace()));
 
         if (deferredSelectionOffset) {
@@ -187,6 +211,7 @@ void KisMask::Private::initSelectionImpl(KisSelectionSP copyFrom, KisLayerSP par
         }
     }
     selection->setParentNode(q);
+    selection->pixelSelection()->setSupportsWraparoundMode(true);
     selection->updateProjection();
 }
 
@@ -197,7 +222,8 @@ KisSelectionSP KisMask::selection() const
 
 KisPaintDeviceSP KisMask::paintDevice() const
 {
-    return selection()->pixelSelection();
+    KisSelectionSP selection = this->selection();
+    return selection ? selection->pixelSelection() : 0;
 }
 
 KisPaintDeviceSP KisMask::original() const
@@ -207,7 +233,15 @@ KisPaintDeviceSP KisMask::original() const
 
 KisPaintDeviceSP KisMask::projection() const
 {
-    return paintDevice();
+    KisPaintDeviceSP originalDevice = original();
+    KisPaintDeviceSP result = originalDevice;
+
+    KisSelectionSP selection = this->selection();
+    if (selection && hasTemporaryTarget()) {
+        result = m_d->safeProjection->getDeviceLazy(selection)->pixelSelection();
+    }
+
+    return result;
 }
 
 KisAbstractProjectionPlaneSP KisMask::projectionPlane() const
@@ -218,11 +252,10 @@ KisAbstractProjectionPlaneSP KisMask::projectionPlane() const
 void KisMask::setSelection(KisSelectionSP selection)
 {
     m_d->selection = selection;
-    if (parent()) {
-        const KisLayer *parentLayer = qobject_cast<const KisLayer*>(parent());
-        m_d->selection->setDefaultBounds(new KisDefaultBounds(parentLayer->image()));
-    }
+    m_d->selection->setDefaultBounds(new KisDefaultBounds(image()));
+    m_d->selection->setResolutionProxy(toQShared(new KisImageResolutionProxy(image())));
     m_d->selection->setParentNode(this);
+    m_d->selection->pixelSelection()->setSupportsWraparoundMode(true);
 }
 
 void KisMask::select(const QRect & rc, quint8 selectedness)
@@ -246,30 +279,39 @@ QRect KisMask::decorateRect(KisPaintDeviceSP &src,
     return rc;
 }
 
+bool KisMask::paintsOutsideSelection() const
+{
+    return false;
+}
+
 void KisMask::apply(KisPaintDeviceSP projection, const QRect &applyRect, const QRect &needRect, PositionToFilthy maskPos) const
 {
     if (selection()) {
 
-        m_d->selection->updateProjection(applyRect);
+        flattenSelectionProjection(m_d->selection, applyRect);
 
         KisSelectionSP effectiveSelection = m_d->selection;
-        QRect effectiveExtent = effectiveSelection->selectedRect();
 
         {
             // Access temporary target under the lock held
             KisIndirectPaintingSupport::ReadLocker l(this);
 
-            if (hasTemporaryTarget()) {
-                effectiveExtent |= temporaryTarget()->extent();
-            }
+            if (!paintsOutsideSelection()) {
+                // extent of m_d->selection should also be accessed under a lock,
+                // because it might be being merged in by the temporary target atm
+                QRect effectiveExtent = m_d->selection->selectedRect();
 
-            if(!effectiveExtent.intersects(applyRect)) {
-                return;
+                if (hasTemporaryTarget()) {
+                    effectiveExtent |= temporaryTarget()->extent();
+                }
+
+                if(!effectiveExtent.intersects(applyRect)) {
+                    return;
+                }
             }
 
             if (hasTemporaryTarget()) {
-                effectiveSelection = m_d->cachedSelection.getSelection();
-                effectiveSelection->setDefaultBounds(m_d->selection->pixelSelection()->defaultBounds());
+                effectiveSelection = m_d->safeProjection->getDeviceLazy(m_d->selection);
 
                 KisPainter::copyAreaOptimized(applyRect.topLeft(),
                                               m_d->selection->pixelSelection(),
@@ -278,63 +320,109 @@ void KisMask::apply(KisPaintDeviceSP projection, const QRect &applyRect, const Q
                 KisPainter gc(effectiveSelection->pixelSelection());
                 setupTemporaryPainter(&gc);
                 gc.bitBlt(applyRect.topLeft(), temporaryTarget(), applyRect);
+            } else {
+                m_d->safeProjection->releaseDevice();
             }
-        }
 
-        KisPaintDeviceSP cacheDevice = m_d->paintDeviceCache.getDevice(projection);
-
-        QRect updatedRect = decorateRect(projection, cacheDevice, applyRect, maskPos);
-
-        // masks don't have any compositioning
-        KisPainter::copyAreaOptimized(updatedRect.topLeft(), cacheDevice, projection, updatedRect, effectiveSelection);
-        m_d->paintDeviceCache.putDevice(cacheDevice);
-
-        if (effectiveSelection != m_d->selection) {
-            m_d->cachedSelection.putSelection(effectiveSelection);
+            mergeInMaskInternal(projection, effectiveSelection, applyRect, needRect, maskPos);
         }
 
     } else {
-        KisPaintDeviceSP cacheDevice = m_d->paintDeviceCache.getDevice(projection);
+        mergeInMaskInternal(projection, 0, applyRect, needRect, maskPos);
+    }
+}
 
-        cacheDevice->makeCloneFromRough(projection, needRect);
-        projection->clear(needRect);
+void KisMask::mergeInMaskInternal(KisPaintDeviceSP projection,
+                                  KisSelectionSP effectiveSelection,
+                                  const QRect &applyRect,
+                                  const QRect &preparedNeedRect,
+                                  KisNode::PositionToFilthy maskPos) const
+{
+    KisCachedPaintDevice::Guard d1(projection, m_d->paintDeviceCache);
+    KisPaintDeviceSP cacheDevice = d1.device();
+
+    if (effectiveSelection) {
+        QRect updatedRect = decorateRect(projection, cacheDevice, applyRect, maskPos);
+
+        // masks don't have any compositing
+        KisPainter::copyAreaOptimized(updatedRect.topLeft(), cacheDevice, projection, updatedRect, effectiveSelection);
+
+    } else {
+        cacheDevice->makeCloneFromRough(projection, preparedNeedRect);
+        projection->clear(preparedNeedRect);
 
         decorateRect(cacheDevice, projection, applyRect, maskPos);
-
-        m_d->paintDeviceCache.putDevice(cacheDevice);
     }
 }
 
+void KisMask::flattenSelectionProjection(KisSelectionSP selection, const QRect &dirtyRect) const
+{
+    selection->updateProjection(dirtyRect);
+}
+
 QRect KisMask::needRect(const QRect &rect,  PositionToFilthy pos) const
 {
     Q_UNUSED(pos);
     QRect resultRect = rect;
-    if (m_d->selection)
-        resultRect &= m_d->selection->selectedRect();
+    if (m_d->selection) {
+        QRect selectionExtent = m_d->selection->selectedRect();
+
+        // copy for thread safety!
+        KisPaintDeviceSP temporaryTarget = this->temporaryTarget();
+
+        if (temporaryTarget) {
+            selectionExtent |= temporaryTarget->extent();
+        }
+
+        resultRect &= selectionExtent;
+    }
 
     return resultRect;
 }
 
 QRect KisMask::changeRect(const QRect &rect, PositionToFilthy pos) const
 {
-    Q_UNUSED(pos);
-    QRect resultRect = rect;
-    if (m_d->selection)
-        resultRect &= m_d->selection->selectedRect();
-
-    return resultRect;
+    return KisMask::needRect(rect, pos);
 }
 
 QRect KisMask::extent() const
 {
-    return m_d->selection ? m_d->selection->selectedRect() :
-           parent() ? parent()->extent() : QRect();
+    QRect resultRect;
+
+    if (m_d->selection) {
+        resultRect = m_d->selection->selectedRect();
+
+        // copy for thread safety!
+        KisPaintDeviceSP temporaryTarget = this->temporaryTarget();
+
+        if (temporaryTarget) {
+            resultRect |= temporaryTarget->extent();
+        }
+    } else if (KisNodeSP parent = this->parent()) {
+        resultRect = parent->extent();
+    }
+
+    return resultRect;
 }
 
 QRect KisMask::exactBounds() const
 {
-    return m_d->selection ? m_d->selection->selectedExactRect() :
-           parent() ? parent()->exactBounds() : QRect();
+    QRect resultRect;
+
+    if (m_d->selection) {
+        resultRect = m_d->selection->selectedExactRect();
+
+        // copy for thread safety!
+        KisPaintDeviceSP temporaryTarget = this->temporaryTarget();
+
+        if (temporaryTarget) {
+            resultRect |= temporaryTarget->exactBounds();
+        }
+    } else if (KisNodeSP parent = this->parent()) {
+        resultRect = parent->exactBounds();
+    }
+
+    return resultRect;
 }
 
 qint32 KisMask::x() const
@@ -378,36 +466,49 @@ QRect KisMask::nonDependentExtent() const
     return QRect();
 }
 
-QImage KisMask::createThumbnail(qint32 w, qint32 h)
+QImage KisMask::createThumbnail(qint32 w, qint32 h, Qt::AspectRatioMode aspectRatioMode)
 {
     KisPaintDeviceSP originalDevice =
         selection() ? selection()->projection() : 0;
 
     return originalDevice ?
-           originalDevice->createThumbnail(w, h, 1,
+           originalDevice->createThumbnail(w, h, aspectRatioMode, 1,
                                            KoColorConversionTransformation::internalRenderingIntent(),
                                            KoColorConversionTransformation::internalConversionFlags()) : QImage();
 }
 
+int KisMask::thumbnailSeqNo() const
+{
+    KisPaintDeviceSP originalDevice =
+        selection() ? selection()->projection() : 0;
+    return originalDevice ? originalDevice->sequenceNumber() : -1;
+}
+
 void KisMask::testingInitSelection(const QRect &rect, KisLayerSP parentLayer)
 {
     if (parentLayer) {
-        m_d->selection = new KisSelection(new KisSelectionDefaultBounds(parentLayer->paintDevice(), parentLayer->image()));
+        m_d->selection = new KisSelection(new KisSelectionDefaultBounds(parentLayer->paintDevice()), toQShared(new KisImageResolutionProxy(image())));
     } else {
-        m_d->selection = new KisSelection();
+        m_d->selection = new KisSelection(new KisSelectionEmptyBounds(), toQShared(new KisImageResolutionProxy(image())));
     }
 
     m_d->selection->pixelSelection()->select(rect, OPACITY_OPAQUE_U8);
+    m_d->selection->pixelSelection()->setSupportsWraparoundMode(true);
     m_d->selection->updateProjection(rect);
     m_d->selection->setParentNode(this);
 }
 
+bool KisMask::supportsLodPainting() const
+{
+    return !m_d->selection || !m_d->selection->hasShapeSelection();
+}
+
 KisKeyframeChannel *KisMask::requestKeyframeChannel(const QString &id)
 {
-    if (id == KisKeyframeChannel::Content.id()) {
+    if (id == KisKeyframeChannel::Raster.id()) {
         KisPaintDeviceSP device = paintDevice();
         if (device) {
-            KisRasterKeyframeChannel *contentChannel = device->createKeyframeChannel(KisKeyframeChannel::Content);
+            KisRasterKeyframeChannel *contentChannel = device->createKeyframeChannel(KisKeyframeChannel::Raster);
             contentChannel->setFilenameSuffix(".pixelselection");
             return contentChannel;
        }
@@ -416,12 +517,21 @@ KisKeyframeChannel *KisMask::requestKeyframeChannel(const QString &id)
     return KisNode::requestKeyframeChannel(id);
 }
 
+bool KisMask::supportsKeyframeChannel(const QString &id)
+{
+    if (id == KisKeyframeChannel::Raster.id() && paintDevice()) {
+        return true;
+    }
+
+    return KisNode::supportsKeyframeChannel(id);
+}
+
 void KisMask::baseNodeChangedCallback()
 {
     KisNodeSP up = parent();
     KisLayer *layer = dynamic_cast<KisLayer*>(up.data());
     if (layer) {
-        layer->notifyChildMaskChanged(this);
+        layer->notifyChildMaskChanged();
     }
     KisNode::baseNodeChangedCallback();
 }
