@@ -1,28 +1,15 @@
 /*
- *  Copyright (C) 2007 Boudewijn Rempt <boud@valdyas.org>
+ *  SPDX-FileCopyrightText: 2007 Boudewijn Rempt <boud@valdyas.org>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "kis_node_manager.h"
 
 #include <QStandardPaths>
 #include <QMessageBox>
-#include <QSignalMapper>
+#include <KisSignalMapper.h>
 #include <QApplication>
-#include <QMessageBox>
 
 #include <kactioncollection.h>
 
@@ -80,6 +67,10 @@
 #include "kis_layer_utils.h"
 #include "krita_utils.h"
 #include "kis_shape_layer.h"
+#include "kis_keyframe_channel.h"
+#include "kis_raster_keyframe_channel.h"
+#include "kis_paint_device_frames_interface.h"
+#include "kis_layer_utils.h"
 
 #include "processing/kis_mirror_processing_visitor.h"
 #include "KisView.h"
@@ -87,6 +78,7 @@
 #include <kis_signals_blocker.h>
 #include <libs/image/kis_layer_properties_icons.h>
 #include <libs/image/commands/kis_node_property_list_command.h>
+#include <KisSynchronizedConnection.h>
 
 struct KisNodeManager::Private {
 
@@ -104,8 +96,8 @@ struct KisNodeManager::Private {
     {
     }
 
-    KisNodeManager * q;
-    KisViewManager * view;
+    KisNodeManager * q {nullptr};
+    KisViewManager * view {nullptr};
     QPointer<KisView>imageView;
     KisLayerManager layerManager;
     KisMaskManager maskManager;
@@ -114,7 +106,7 @@ struct KisNodeManager::Private {
     QScopedPointer<KisNodeInsertionAdapter> nodeInsertionAdapter;
     QScopedPointer<KisNodeDisplayModeAdapter> nodeDisplayModeAdapter;
 
-    KisAction *showInTimeline;
+    KisAction *pinToTimeline {nullptr};
 
     KisNodeList selectedNodes;
     QPointer<KisNodeJugglerCompressed> nodeJuggler;
@@ -123,10 +115,11 @@ struct KisNodeManager::Private {
 
     bool activateNodeImpl(KisNodeSP node);
 
-    QSignalMapper nodeCreationSignalMapper;
-    QSignalMapper nodeConversionSignalMapper;
+    KisSignalMapper nodeCreationSignalMapper;
+    KisSignalMapper nodeConversionSignalMapper;
 
-    bool lastRequestedIsolatedModeStatus;
+    bool lastRequestedIsolatedModeStatus {false};
+    KisSynchronizedConnection<KisNodeSP, KisNodeList> activateNodeConnection;
 
     void saveDeviceAsImage(KisPaintDeviceSP device,
                            const QString &defaultName,
@@ -197,8 +190,7 @@ bool KisNodeManager::Private::activateNodeImpl(KisNodeSP node)
 KisNodeManager::KisNodeManager(KisViewManager *view)
     : m_d(new Private(this, view))
 {
-
-    connect(&m_d->layerManager, SIGNAL(sigLayerActivated(KisLayerSP)), SIGNAL(sigLayerActivated(KisLayerSP)));
+    m_d->activateNodeConnection.connectOutputSlot(this, &KisNodeManager::slotImageRequestNodeReselection);
 }
 
 KisNodeManager::~KisNodeManager()
@@ -216,6 +208,7 @@ void KisNodeManager::setView(QPointer<KisView>imageView)
         Q_ASSERT(shapeController);
         shapeController->disconnect(SIGNAL(sigActivateNode(KisNodeSP)), this);
         m_d->imageView->image()->disconnect(this);
+        m_d->imageView->image()->disconnect(&m_d->activateNodeConnection);
     }
 
     m_d->imageView = imageView;
@@ -224,9 +217,9 @@ void KisNodeManager::setView(QPointer<KisView>imageView)
         KisShapeController *shapeController = dynamic_cast<KisShapeController*>(m_d->imageView->document()->shapeController());
         Q_ASSERT(shapeController);
         connect(shapeController, SIGNAL(sigActivateNode(KisNodeSP)), SLOT(slotNonUiActivatedNode(KisNodeSP)));
-        connect(m_d->imageView->image(), SIGNAL(sigIsolatedModeChanged()),this, SLOT(slotUpdateIsolateModeActionImageStatusChange()));
-        connect(m_d->imageView->image(), SIGNAL(sigRequestNodeReselection(KisNodeSP,KisNodeList)),this, SLOT(slotImageRequestNodeReselection(KisNodeSP,KisNodeList)));
+        m_d->activateNodeConnection.connectInputSignal(m_d->imageView->image(), &KisImage::sigRequestNodeReselection);
         m_d->imageView->resourceProvider()->slotNodeActivated(m_d->imageView->currentNode());
+        connect(m_d->imageView->image(), SIGNAL(sigIsolatedModeChanged()), this, SLOT(handleExternalIsolationChange()));
     }
 
 }
@@ -252,7 +245,7 @@ void KisNodeManager::setView(QPointer<KisView>imageView)
 #define CONVERT_NODE_ACTION(id, layerType)              \
     CONVERT_NODE_ACTION_2(id, layerType, layerType)
 
-void KisNodeManager::setup(KActionCollection * actionCollection, KisActionManager* actionManager)
+void KisNodeManager::setup(KisKActionCollection * actionCollection, KisActionManager* actionManager)
 {
     m_d->layerManager.setup(actionManager);
     m_d->maskManager.setup(actionCollection, actionManager);
@@ -274,9 +267,15 @@ void KisNodeManager::setup(KActionCollection * actionCollection, KisActionManage
     action = actionManager->createAction("activateNextLayer");
     connect(action, SIGNAL(triggered()), this, SLOT(activateNextNode()));
 
+    action = actionManager->createAction("activateNextSiblingLayer");
+    connect(action, SIGNAL(triggered()), this, SLOT(activateNextSiblingNode()));
+
     action = actionManager->createAction("activatePreviousLayer");
     connect(action, SIGNAL(triggered()), this, SLOT(activatePreviousNode()));
 
+    action = actionManager->createAction("activatePreviousSiblingLayer");
+    connect(action, SIGNAL(triggered()), this, SLOT(activatePreviousSiblingNode()));
+
     action = actionManager->createAction("switchToPreviouslyActiveNode");
     connect(action, SIGNAL(triggered()), this, SLOT(switchToPreviouslyActiveNode()));
 
@@ -326,10 +325,10 @@ void KisNodeManager::setup(KActionCollection * actionCollection, KisActionManage
     action = actionManager->createAction("new_from_visible");
     connect(action, SIGNAL(triggered()), this, SLOT(createFromVisible()));
 
-    action = actionManager->createAction("show_in_timeline");
+    action = actionManager->createAction("pin_to_timeline");
     action->setCheckable(true);
-    connect(action, SIGNAL(toggled(bool)), this, SLOT(slotShowHideTimeline(bool)));
-    m_d->showInTimeline = action;
+    connect(action, SIGNAL(toggled(bool)), this, SLOT(slotPinToTimeline(bool)));
+    m_d->pinToTimeline = action;
 
     NEW_LAYER_ACTION("add_new_paint_layer", "KisPaintLayer");
 
@@ -373,8 +372,13 @@ void KisNodeManager::setup(KActionCollection * actionCollection, KisActionManage
     connect(&m_d->nodeConversionSignalMapper, SIGNAL(mapped(QString)),
             this, SLOT(convertNode(QString)));
 
-    action = actionManager->createAction("isolate_layer");
-    connect(action, SIGNAL(triggered(bool)), this, SLOT(toggleIsolateMode(bool)));
+    // Isolation Modes...
+    // Post Qt5.14 this can be replaced with QActionGroup + ExclusionPolicy::ExclusiveOptional.
+    action = actionManager->createAction("isolate_active_layer");
+    connect(action, SIGNAL(toggled(bool)), this, SLOT(setIsolateActiveLayerMode(bool)));
+    action = actionManager->createAction("isolate_active_group");
+    connect(action, SIGNAL(triggered(bool)), this, SLOT(setIsolateActiveGroupMode(bool)));
+    connect(this, SIGNAL(sigNodeActivated(KisNodeSP)), SLOT(changeIsolationRoot(KisNodeSP)));
 
     action = actionManager->createAction("toggle_layer_visibility");
     connect(action, SIGNAL(triggered()), this, SLOT(toggleVisibility()));
@@ -397,9 +401,6 @@ void KisNodeManager::setup(KActionCollection * actionCollection, KisActionManage
     // HINT: we can save even when the nodes are not editable
     action  = actionManager->createAction("split_alpha_save_merged");
     connect(action, SIGNAL(triggered()), this, SLOT(slotSplitAlphaSaveMerged()));
-
-    connect(this, SIGNAL(sigNodeActivated(KisNodeSP)), SLOT(slotUpdateIsolateModeAction()));
-    connect(this, SIGNAL(sigNodeActivated(KisNodeSP)), SLOT(slotTryRestartIsolatedMode()));
 }
 
 void KisNodeManager::updateGUI()
@@ -435,6 +436,59 @@ const KoColorSpace* KisNodeManager::activeColorSpace()
     }
 }
 
+bool KisNodeManager::canModifyLayers(KisNodeList nodes, bool showWarning)
+{
+    KisNodeSP lockedNode;
+    Q_FOREACH (KisNodeSP node, nodes) {
+        if (!node->isEditable(false)) {
+            lockedNode = node;
+            break;
+        }
+    }
+
+    if (lockedNode && showWarning) {
+        QString errorMessage;
+
+        if (nodes.size() <= 1) {
+            errorMessage = i18n("Layer is locked");
+        } else {
+            errorMessage = i18n("Layer \"%1\" is locked", lockedNode->name());
+        }
+
+        m_d->view->showFloatingMessage(errorMessage, QIcon());
+    }
+
+    return !lockedNode;
+}
+
+bool KisNodeManager::canModifyLayer(KisNodeSP node, bool showWarning)
+{
+    return canModifyLayers({node}, showWarning);
+}
+
+bool KisNodeManager::canMoveLayers(KisNodeList nodes, bool showWarning)
+{
+    KisNodeSP lockedNode;
+    Q_FOREACH (KisNodeSP node, nodes) {
+        if (node->parent() && !node->parent()->isEditable(false)) {
+            lockedNode = node->parent();
+            break;
+        }
+    }
+
+    if (lockedNode && showWarning) {
+        QString errorMessage = i18n("Layer \"%1\" is locked", lockedNode->name());
+        m_d->view->showFloatingMessage(errorMessage, QIcon());
+    }
+
+    return !lockedNode;
+}
+
+bool KisNodeManager::canMoveLayer(KisNodeSP node, bool showWarning)
+{
+    return canMoveLayers({node}, showWarning);
+}
+
 void KisNodeManager::moveNodeAt(KisNodeSP node, KisNodeSP parent, int index)
 {
     if (parent->allowAsChild(node)) {
@@ -472,74 +526,77 @@ void KisNodeManager::addNodesDirect(KisNodeList nodes, KisNodeSP parent, KisNode
 
 void KisNodeManager::toggleIsolateActiveNode()
 {
-    KisImageWSP image = m_d->view->image();
-    KisNodeSP activeNode = this->activeNode();
-    KIS_ASSERT_RECOVER_RETURN(activeNode);
-
-    if (activeNode == image->isolatedModeRoot()) {
-        toggleIsolateMode(false);
-    } else {
-        toggleIsolateMode(true);
-    }
+    QAction* action = m_d->view->actionManager()->actionByName("isolate_active_layer");
+    action->toggle();
 }
 
-void KisNodeManager::toggleIsolateMode(bool checked)
+void KisNodeManager::setIsolateActiveLayerMode(bool checked)
 {
     KisImageWSP image = m_d->view->image();
+    KIS_ASSERT_RECOVER_RETURN(image);
 
+    const bool groupIsolationState = image->isIsolatingGroup();
+    changeIsolationMode(checked, groupIsolationState);
+}
+
+void KisNodeManager::setIsolateActiveGroupMode(bool checked)
+{
+    KisImageWSP image = m_d->view->image();
+    KIS_ASSERT_RECOVER_RETURN(image);
+
+    const bool layerIsolationState = image->isIsolatingLayer();
+    changeIsolationMode(layerIsolationState, checked);
+}
+
+void KisNodeManager::changeIsolationMode(bool isolateActiveLayer, bool isolateActiveGroup)
+{
+    KisImageWSP image = m_d->view->image();
     KisNodeSP activeNode = this->activeNode();
-    if (checked && activeNode) {
+    KIS_ASSERT_RECOVER_RETURN(image && activeNode);
 
-        // Transform and colorize masks don't have pixel data...
-        if (activeNode->inherits("KisTransformMask") ||
-                activeNode->inherits("KisColorizeMask")) return;
-
-        if (!image->startIsolatedMode(activeNode)) {
-            KisAction *action = m_d->view->actionManager()->actionByName("isolate_layer");
-            action->setChecked(false);
+    if (isolateActiveLayer || isolateActiveGroup) {
+        if (image->startIsolatedMode(activeNode, isolateActiveLayer, isolateActiveGroup) == false) {
+            reinitializeIsolationActionGroup();
         }
     } else {
         image->stopIsolatedMode();
     }
-
-    m_d->lastRequestedIsolatedModeStatus = checked;
 }
 
-void KisNodeManager::slotUpdateIsolateModeActionImageStatusChange()
+void KisNodeManager::changeIsolationRoot(KisNodeSP isolationRoot)
 {
-    slotUpdateIsolateModeAction();
+    KisImageWSP image = m_d->view->image();
+    if (!image || !isolationRoot) return;
 
-    KisNodeSP isolatedRootNode = m_d->view->image()->isolatedModeRoot();
-    if (this->activeNode() &&
-            bool(isolatedRootNode) != m_d->lastRequestedIsolatedModeStatus) {
+    const bool isIsolatingLayer = image->isIsolatingLayer();
+    const bool isIsolatingGroup = image->isIsolatingGroup();
 
-        slotTryRestartIsolatedMode();
+    // Restart isolation with a new root node and the same settings.
+    if (image->startIsolatedMode(isolationRoot, isIsolatingLayer, isIsolatingGroup) == false) {
+        reinitializeIsolationActionGroup();
     }
 }
 
-void KisNodeManager::slotUpdateIsolateModeAction()
+void KisNodeManager::handleExternalIsolationChange()
 {
-    KisAction *action = m_d->view->actionManager()->actionByName("isolate_layer");
-    Q_ASSERT(action);
+    // It might be that we have multiple Krita windows open. In such a case
+    // only the currently active one should restart isolated mode
+    if (!m_d->view->mainWindowAsQWidget()->isActiveWindow()) return;
 
+    KisImageWSP image = m_d->view->image();
     KisNodeSP activeNode = this->activeNode();
-    KisNodeSP isolatedRootNode = m_d->view->image()->isolatedModeRoot();
 
-    action->setChecked(isolatedRootNode && isolatedRootNode == activeNode);
+    const bool isIsolatingLayer = image->isIsolatingLayer();
+    const bool isIsolatingGroup = image->isIsolatingGroup();
+
+    m_d->view->actionManager()->actionByName("isolate_active_layer")->setChecked(isIsolatingLayer);
+    m_d->view->actionManager()->actionByName("isolate_active_group")->setChecked(isIsolatingGroup);
 }
 
-void KisNodeManager::slotTryRestartIsolatedMode()
+void KisNodeManager::reinitializeIsolationActionGroup()
 {
-    /**
-     * It might be that we have multiple Krita windows open. In such a case
-     * only the currently active one should restart isolated mode
-     */
-    if (!m_d->view->mainWindow()->isActiveWindow()) return;
-
-    KisNodeSP isolatedRootNode = m_d->view->image()->isolatedModeRoot();
-    if (!isolatedRootNode && !m_d->lastRequestedIsolatedModeStatus) return;
-
-    this->toggleIsolateMode(true);
+    m_d->view->actionManager()->actionByName("isolate_active_layer")->setChecked(false);
+    m_d->view->actionManager()->actionByName("isolate_active_group")->setChecked(false);
 }
 
 KisNodeSP  KisNodeManager::createNode(const QString & nodeType, bool quiet, KisPaintDeviceSP copyFrom)
@@ -555,6 +612,10 @@ KisNodeSP  KisNodeManager::createNode(const QString & nodeType, bool quiet, KisP
 
     KIS_ASSERT_RECOVER_RETURN_VALUE(activeNode, 0);
 
+    /// the check for editability happens inside the functions
+    /// themselves, because layers can be created anyway (in a
+    /// different position), but masks cannot.
+
     // XXX: make factories for this kind of stuff,
     //      with a registry
 
@@ -596,10 +657,10 @@ void KisNodeManager::createFromVisible()
     KisLayerUtils::newLayerFromVisible(m_d->view->image(), m_d->view->image()->root()->lastChild());
 }
 
-void KisNodeManager::slotShowHideTimeline(bool value)
+void KisNodeManager::slotPinToTimeline(bool value)
 {
     Q_FOREACH (KisNodeSP node, selectedNodes()) {
-        node->setUseInTimeline(value);
+        node->setPinnedToTimeline(value);
     }
 }
 
@@ -618,6 +679,8 @@ void KisNodeManager::convertNode(const QString &nodeType)
     KisNodeSP activeNode = this->activeNode();
     if (!activeNode) return;
 
+    if (!canModifyLayer(activeNode)) return;
+
     if (nodeType == "KisPaintLayer") {
         m_d->layerManager.convertNodeToPaintLayer(activeNode);
     } else if (nodeType == "KisSelectionMask" ||
@@ -708,10 +771,8 @@ void KisNodeManager::slotUiActivatedNode(KisNodeSP node)
     if (node) {
         QStringList vectorTools = QStringList()
                 << "InteractionTool"
-                << "KarbonPatternTool"
                 << "KarbonGradientTool"
                 << "KarbonCalligraphyTool"
-                << "CreateShapesTool"
                 << "PathTool";
 
         QStringList pixelTools = QStringList()
@@ -755,8 +816,8 @@ void KisNodeManager::nodesUpdated()
     m_d->view->selectionManager()->selectionChanged();
 
     {
-        KisSignalsBlocker b(m_d->showInTimeline);
-        m_d->showInTimeline->setChecked(node->useInTimeline());
+        KisSignalsBlocker b(m_d->pinToTimeline);
+        m_d->pinToTimeline->setChecked(node->isPinnedToTimeline());
     }
 }
 
@@ -883,7 +944,7 @@ bool KisNodeManager::trySetNodeProperties(KisNodeSP node, KisImageSP image, KisB
         }
     }
 
-    KisNodePropertyListCommand::setNodePropertiesNoUndo(node, image, properties);
+    KisNodePropertyListCommand::setNodePropertiesAutoUndo(node, image, properties);
 
     return true;
 }
@@ -927,6 +988,8 @@ KisNodeJugglerCompressed* KisNodeManager::Private::lazyGetJuggler(const KUndo2Ma
 
 void KisNodeManager::raiseNode()
 {
+    if (!canMoveLayers(selectedNodes())) return;
+
     KUndo2MagicString actionName = kundo2_i18n("Raise Nodes");
     KisNodeJugglerCompressed *juggler = m_d->lazyGetJuggler(actionName);
     juggler->raiseNode(selectedNodes());
@@ -934,6 +997,8 @@ void KisNodeManager::raiseNode()
 
 void KisNodeManager::lowerNode()
 {
+    if (!canMoveLayers(selectedNodes())) return;
+
     KUndo2MagicString actionName = kundo2_i18n("Lower Nodes");
     KisNodeJugglerCompressed *juggler = m_d->lazyGetJuggler(actionName);
     juggler->lowerNode(selectedNodes());
@@ -952,6 +1017,8 @@ void KisNodeManager::removeSingleNode(KisNodeSP node)
 
 void KisNodeManager::removeSelectedNodes(KisNodeList nodes)
 {
+    if (!canModifyLayers(nodes)) return;
+
     KUndo2MagicString actionName = kundo2_i18n("Remove Nodes");
     KisNodeJugglerCompressed *juggler = m_d->lazyGetJuggler(actionName);
     juggler->removeNode(nodes);
@@ -1002,56 +1069,74 @@ void KisNodeManager::mirrorAllNodesY()
                Qt::Vertical, m_d->view->selection());
 }
 
-void KisNodeManager::activateNextNode()
+void KisNodeManager::activateNextNode(bool siblingsOnly)
 {
     KisNodeSP activeNode = this->activeNode();
     if (!activeNode) return;
 
-    KisNodeSP node = activeNode->nextSibling();
+    KisNodeSP nextNode = activeNode->nextSibling();
 
-    while (node && node->childCount() > 0) {
-        node = node->firstChild();
+    if (!siblingsOnly) {
+        // Recurse groups...
+        while (nextNode && nextNode->childCount() > 0) {
+            nextNode = nextNode->firstChild();
+        }
+
+        // Out of nodes? Back out of group...
+        if (!nextNode && activeNode->parent()) {
+            nextNode = activeNode->parent();
+        }
     }
 
-    if (!node && activeNode->parent() && activeNode->parent()->parent()) {
-        node = activeNode->parent();
+    // Skip nodes hidden from tree view..
+    while (nextNode && isNodeHidden(nextNode, m_d->nodeDisplayModeAdapter->showGlobalSelectionMask())) {
+        nextNode = nextNode->nextSibling();
     }
 
-    while(node && isNodeHidden(node, true)) {
-        node = node->nextSibling();
-    }
-
-    if (node) {
-        slotNonUiActivatedNode(node);
+    // Select node, unless root..
+    if (nextNode && nextNode->parent()) {
+        slotNonUiActivatedNode(nextNode);
     }
 }
 
-void KisNodeManager::activatePreviousNode()
+void KisNodeManager::activateNextSiblingNode()
+{
+    activateNextNode(true);
+}
+
+void KisNodeManager::activatePreviousNode(bool siblingsOnly)
 {
     KisNodeSP activeNode = this->activeNode();
     if (!activeNode) return;
 
-    KisNodeSP node;
+    KisNodeSP nextNode = activeNode->prevSibling();
 
-    if (activeNode->childCount() > 0) {
-        node = activeNode->lastChild();
-    }
-    else {
-        node = activeNode->prevSibling();
+    if (!siblingsOnly) {
+        // Enter groups..
+        if (activeNode->childCount() > 0) {
+            nextNode = activeNode->lastChild();
+        }
+
+        // Out of nodes? Back out of group...
+        if (!nextNode && activeNode->parent()) {
+            nextNode = activeNode->parent()->prevSibling();
+        }
     }
 
-    while (!node && activeNode->parent()) {
-        node = activeNode->parent()->prevSibling();
-        activeNode = activeNode->parent();
+    // Skip nodes hidden from tree view..
+    while (nextNode && isNodeHidden(nextNode, m_d->nodeDisplayModeAdapter->showGlobalSelectionMask())) {
+        nextNode = nextNode->prevSibling();
     }
 
-    while(node && isNodeHidden(node, true)) {
-        node = node->prevSibling();
+    // Select node, unless root..
+    if (nextNode && nextNode->parent()) {
+        slotNonUiActivatedNode(nextNode);
     }
+}
 
-    if (node) {
-        slotNonUiActivatedNode(node);
-    }
+void KisNodeManager::activatePreviousSiblingNode()
+{
+    activatePreviousNode(true);
 }
 
 void KisNodeManager::switchToPreviouslyActiveNode()
@@ -1066,8 +1151,9 @@ void KisNodeManager::mirrorNode(KisNodeSP node,
                                 Qt::Orientation orientation,
                                 KisSelectionSP selection)
 {
+    if (!canModifyLayer(node)) return;
+
     KisImageSignalVector emitSignals;
-    emitSignals << ModifiedSignal;
 
     KisProcessingApplicator applicator(m_d->view->image(), node,
                                        KisProcessingApplicator::RECURSIVE,
@@ -1099,7 +1185,7 @@ void KisNodeManager::Private::saveDeviceAsImage(KisPaintDeviceSP device,
                                                 qreal yRes,
                                                 quint8 opacity)
 {
-    KoFileDialog dialog(view->mainWindow(), KoFileDialog::SaveFile, "savenodeasimage");
+    KoFileDialog dialog(view->mainWindowAsQWidget(), KoFileDialog::SaveFile, "savenodeasimage");
     dialog.setCaption(i18n("Export \"%1\"", defaultName));
     dialog.setDefaultDir(QStandardPaths::writableLocation(QStandardPaths::PicturesLocation));
     dialog.setMimeTypeFilters(KisImportExportManager::supportedMimeTypes(KisImportExportManager::Export));
@@ -1107,9 +1193,8 @@ void KisNodeManager::Private::saveDeviceAsImage(KisPaintDeviceSP device,
 
     if (filename.isEmpty()) return;
 
-    QUrl url = QUrl::fromLocalFile(filename);
 
-    if (url.isEmpty()) return;
+    if (filename.isEmpty()) return;
 
     QString mimefilter = KisMimeDatabase::mimeTypeForFile(filename, false);
 
@@ -1128,8 +1213,8 @@ void KisNodeManager::Private::saveDeviceAsImage(KisPaintDeviceSP device,
 
     dst->initialRefreshGraph();
 
-    if (!doc->exportDocumentSync(url, mimefilter.toLatin1())) {
-        QMessageBox::warning(0,
+    if (!doc->exportDocumentSync(filename, mimefilter.toLatin1())) {
+        QMessageBox::warning(qApp->activeWindow(),
                              i18nc("@title:window", "Krita"),
                              i18n("Could not save the layer. %1", doc->errorMessage().toUtf8().data()),
                              QMessageBox::Ok);
@@ -1146,10 +1231,17 @@ void KisNodeManager::saveNodeAsImage()
         return;
     }
 
-    KisImageWSP image = m_d->view->image();
+    KisPaintDeviceSP saveDevice = node->projection();
+
+    if (!saveDevice) {
+        m_d->view->showFloatingMessage(i18nc("warning message when trying to export a transform mask", "Layer has no pixel data"), QIcon());
+        return;
+    }
+
+    KisImageSP image = m_d->view->image();
     QRect saveRect = image->bounds() | node->exactBounds();
 
-    m_d->saveDeviceAsImage(node->projection(),
+    m_d->saveDeviceAsImage(saveDevice,
                            node->name(),
                            saveRect,
                            image->xRes(), image->yRes(),
@@ -1165,7 +1257,7 @@ void KisNodeManager::saveVectorLayerAsImage()
         return;
     }
 
-    KoFileDialog dialog(m_d->view->mainWindow(), KoFileDialog::SaveFile, "savenodeasimage");
+    KoFileDialog dialog(m_d->view->mainWindowAsQWidget(), KoFileDialog::SaveFile, "savenodeasimage");
     dialog.setCaption(i18nc("@title:window", "Export to SVG"));
     dialog.setDefaultDir(QStandardPaths::writableLocation(QStandardPaths::PicturesLocation));
     dialog.setMimeTypeFilters(QStringList() << "image/svg+xml", "image/svg+xml");
@@ -1192,38 +1284,14 @@ void KisNodeManager::saveVectorLayerAsImage()
 
 void KisNodeManager::slotSplitAlphaIntoMask()
 {
+
     KisNodeSP node = activeNode();
+    if (!canModifyLayer(node)) return;
 
     // guaranteed by KisActionManager
     KIS_ASSERT_RECOVER_RETURN(node->hasEditablePaintDevice());
 
-    KisPaintDeviceSP srcDevice = node->paintDevice();
-    const KoColorSpace *srcCS = srcDevice->colorSpace();
-    const QRect processRect =
-            srcDevice->exactBounds() |
-            srcDevice->defaultBounds()->bounds();
-
-    KisPaintDeviceSP selectionDevice =
-            new KisPaintDevice(KoColorSpaceRegistry::instance()->alpha8());
-
-    m_d->commandsAdapter.beginMacro(kundo2_i18n("Split Alpha into a Mask"));
-    KisTransaction transaction(kundo2_noi18n("__split_alpha_channel__"), srcDevice);
-
-    KisSequentialIterator srcIt(srcDevice, processRect);
-    KisSequentialIterator dstIt(selectionDevice, processRect);
-
-    while (srcIt.nextPixel() && dstIt.nextPixel()) {
-        quint8 *srcPtr = srcIt.rawData();
-        quint8 *alpha8Ptr = dstIt.rawData();
-
-        *alpha8Ptr = srcCS->opacityU8(srcPtr);
-        srcCS->setOpacity(srcPtr, OPACITY_OPAQUE_U8, 1);
-    }
-
-    m_d->commandsAdapter.addExtraCommand(transaction.endAndTake());
-
-    createNode("KisTransparencyMask", false, selectionDevice);
-    m_d->commandsAdapter.endMacro();
+    KisLayerUtils::splitAlphaToMask(node->image(), node, m_d->maskManager.createMaskNameCommon(node, "KisTransparencyMask",  i18n("Transparency Mask")));
 }
 
 void KisNodeManager::Private::mergeTransparencyMaskAsAlpha(bool writeToLayers)
@@ -1234,8 +1302,8 @@ void KisNodeManager::Private::mergeTransparencyMaskAsAlpha(bool writeToLayers)
     // guaranteed by KisActionManager
     KIS_ASSERT_RECOVER_RETURN(node->inherits("KisTransparencyMask"));
 
-    if (writeToLayers && !parentNode->hasEditablePaintDevice()) {
-        QMessageBox::information(view->mainWindow(),
+    if (writeToLayers && (!parentNode->hasEditablePaintDevice() || !node->isEditable(false))) {
+        QMessageBox::information(view->mainWindowAsQWidget(),
                                  i18nc("@title:window", "Layer %1 is not editable", parentNode->name()),
                                  i18n("Cannot write alpha channel of "
                                       "the parent layer \"%1\".\n"
@@ -1317,7 +1385,7 @@ void KisNodeManager::toggleLock()
     bool isLocked = active->userLocked();
 
     for (auto &node : nodes) {
-        node->setUserLocked(!isLocked);
+        KisLayerPropertiesIcons::setNodePropertyAutoUndo(node, KisLayerPropertiesIcons::locked, !isLocked, m_d->view->image());
     }
 }
 
@@ -1330,8 +1398,7 @@ void KisNodeManager::toggleVisibility()
     bool isVisible = active->visible();
 
     for (auto &node : nodes) {
-        node->setVisible(!isVisible);
-        node->setDirty();
+        KisLayerPropertiesIcons::setNodePropertyAutoUndo(node, KisLayerPropertiesIcons::visible, !isVisible, m_d->view->image());
     }
 }
 
@@ -1350,7 +1417,7 @@ void KisNodeManager::toggleAlphaLock()
     for (auto &node : nodes) {
         auto layer = qobject_cast<KisPaintLayer*>(node.data());
         if (layer) {
-            layer->setAlphaLocked(!isAlphaLocked);
+            KisLayerPropertiesIcons::setNodePropertyAutoUndo(node, KisLayerPropertiesIcons::alphaLocked, !isAlphaLocked, m_d->view->image());
         }
     }
 }
@@ -1370,8 +1437,7 @@ void KisNodeManager::toggleInheritAlpha()
     for (auto &node : nodes) {
         auto layer = qobject_cast<KisLayer*>(node.data());
         if (layer) {
-            layer->disableAlphaChannel(!isAlphaDisabled);
-            node->setDirty();
+            KisLayerPropertiesIcons::setNodePropertyAutoUndo(node, KisLayerPropertiesIcons::inheritAlpha, !isAlphaDisabled, m_d->view->image());
         }
     }
 }
@@ -1381,11 +1447,22 @@ void KisNodeManager::cutLayersToClipboard()
     KisNodeList nodes = this->selectedNodes();
     if (nodes.isEmpty()) return;
 
+    KisNodeList::Iterator it = nodes.begin();
+    while (it != nodes.end()) {
+        // make sure the deleted nodes aren't referenced here again
+        if (!it->data()->parent()) {
+            nodes.erase(it);
+        }
+        it++;
+    }
+
     KisClipboard::instance()->setLayers(nodes, m_d->view->image(), false);
 
-    KUndo2MagicString actionName = kundo2_i18n("Cut Nodes");
-    KisNodeJugglerCompressed *juggler = m_d->lazyGetJuggler(actionName);
-    juggler->removeNode(nodes);
+    if (canModifyLayers(nodes)) {
+        KUndo2MagicString actionName = kundo2_i18n("Cut Nodes");
+        KisNodeJugglerCompressed *juggler = m_d->lazyGetJuggler(actionName);
+        juggler->removeNode(nodes);
+    }
 }
 
 void KisNodeManager::copyLayersToClipboard()
@@ -1394,7 +1471,7 @@ void KisNodeManager::copyLayersToClipboard()
     KisClipboard::instance()->setLayers(nodes, m_d->view->image(), true);
 }
 
-void KisNodeManager::pasteLayersFromClipboard()
+void KisNodeManager::pasteLayersFromClipboard(bool changeOffset, QPointF offset)
 {
     const QMimeData *data = KisClipboard::instance()->layersMimeData();
     if (!data) return;
@@ -1418,29 +1495,33 @@ void KisNodeManager::pasteLayersFromClipboard()
                                   parentDummy,
                                   aboveThisDummy,
                                   copyNode,
-                                  nodeInsertionAdapter());
+                                  nodeInsertionAdapter(),
+                                  changeOffset,
+                                  offset);
 }
 
-void KisNodeManager::createQuickGroupImpl(KisNodeJugglerCompressed *juggler,
+bool KisNodeManager::createQuickGroupImpl(KisNodeJugglerCompressed *juggler,
                                           const QString &overrideGroupName,
                                           KisNodeSP *newGroup,
                                           KisNodeSP *newLastChild)
 {
     KisNodeSP active = activeNode();
-    if (!active) return;
+    if (!active) return false;
+
+    if (!canMoveLayer(active)) return false;
 
     KisImageSP image = m_d->view->image();
-    QString groupName = !overrideGroupName.isEmpty() ? overrideGroupName : image->nextLayerName();
+    QString groupName = !overrideGroupName.isEmpty() ? overrideGroupName : image->nextLayerName(i18nc("A group of layers", "Group"));
     KisGroupLayerSP group = new KisGroupLayer(image.data(), groupName, OPACITY_OPAQUE_U8);
 
     KisNodeList nodes1;
     nodes1 << group;
 
     KisNodeList nodes2;
-    nodes2 = KisLayerUtils::sortMergableNodes(image->root(), selectedNodes());
-    KisLayerUtils::filterMergableNodes(nodes2);
+    nodes2 = KisLayerUtils::sortMergeableNodes(image->root(), selectedNodes());
+    KisLayerUtils::filterMergeableNodes(nodes2);
 
-    if (nodes2.size() == 0) return;
+    if (nodes2.size() == 0) return false;
 
     if (KisLayerUtils::checkIsChildOf(active, nodes2)) {
         active = nodes2.first();
@@ -1454,6 +1535,8 @@ void KisNodeManager::createQuickGroupImpl(KisNodeJugglerCompressed *juggler,
 
     *newGroup = group;
     *newLastChild = nodes2.last();
+
+    return true;
 }
 
 void KisNodeManager::createQuickGroup()
@@ -1476,12 +1559,12 @@ void KisNodeManager::createQuickClippingGroup()
     KisNodeSP above;
 
     KisImageSP image = m_d->view->image();
-    createQuickGroupImpl(juggler, image->nextLayerName(i18nc("default name for a clipping group layer", "Clipping Group")), &parent, &above);
+    if (createQuickGroupImpl(juggler, image->nextLayerName(i18nc("default name for a clipping group layer", "Clipping Group")), &parent, &above)) {
+        KisPaintLayerSP maskLayer = new KisPaintLayer(image.data(), i18nc("default name for quick clip group mask layer", "Mask Layer"), OPACITY_OPAQUE_U8, image->colorSpace());
+        maskLayer->disableAlphaChannel(true);
 
-    KisPaintLayerSP maskLayer = new KisPaintLayer(image.data(), i18nc("default name for quick clip group mask layer", "Mask Layer"), OPACITY_OPAQUE_U8, image->colorSpace());
-    maskLayer->disableAlphaChannel(true);
-
-    juggler->addNode(KisNodeList() << maskLayer, parent, above);
+        juggler->addNode(KisNodeList() << maskLayer, parent, above);
+    }
 }
 
 void KisNodeManager::quickUngroup()
@@ -1489,6 +1572,8 @@ void KisNodeManager::quickUngroup()
     KisNodeSP active = activeNode();
     if (!active) return;
 
+    if (!canModifyLayer(active)) return;
+
     KisNodeSP parent = active->parent();
     KisNodeSP aboveThis = active;
 
@@ -1581,3 +1666,14 @@ void KisNodeManager::selectUnlockedNodes()
 
     selectLayersImpl(props, invertedProps);
 }
+
+void KisNodeManager::slotUiActivateNode()
+{
+    if (!sender()->property("node").isNull()) {
+        QString name = sender()->property("node").toString();
+        KisNodeSP node = KisLayerUtils::findNodeByName(m_d->imageView->image()->rootLayer(),name);
+        if (node) {
+            slotUiActivatedNode(node);
+        }
+    }
+}
