@@ -1,20 +1,8 @@
 /*
- *  Copyright (c) 2009 Cyrille Berger <cberger@cberger.net>
- *  Copyright (c) 2017 Scott Petrovic <scottpetrovic@gmail.com>
+ *  SPDX-FileCopyrightText: 2009 Cyrille Berger <cberger@cberger.net>
+ *  SPDX-FileCopyrightText: 2017 Scott Petrovic <scottpetrovic@gmail.com>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "kis_painting_assistants_decoration.h"
@@ -29,10 +17,14 @@
 #include "kis_debug.h"
 #include "KisDocument.h"
 #include "kis_canvas2.h"
+#include "kis_canvas_resource_provider.h"
 #include "kis_icon_utils.h"
 #include "KisViewManager.h"
+#include <KoCompositeOpRegistry.h>
+#include "kis_tool_proxy.h"
 
 #include <QPainter>
+#include <QPainterPath>
 #include <QApplication>
 
 struct KisPaintingAssistantsDecoration::Private {
@@ -40,6 +32,7 @@ struct KisPaintingAssistantsDecoration::Private {
         : assistantVisible(false)
         , outlineVisible(false)
         , snapOnlyOneAssistant(true)
+        , snapEraser(false)
         , firstAssistant(0)
         , aFirstStroke(false)
         , m_handleSize(14)
@@ -48,22 +41,27 @@ struct KisPaintingAssistantsDecoration::Private {
     bool assistantVisible;
     bool outlineVisible;
     bool snapOnlyOneAssistant;
+    bool snapEraser;
     KisPaintingAssistantSP firstAssistant;
     KisPaintingAssistantSP selectedAssistant;
     bool aFirstStroke;
     bool m_isEditingAssistants = false;
-    bool m_outlineVisible = false;
     int m_handleSize; // size of editor handles on assistants
 
     // move, visibility, delete icons for each assistant. These only display while the assistant tool is active
-    // these icons will be covered by the kis_paintint_assistant_decoration with things like the perspective assistant
+    // these icons will be covered by the kis_painting_assistant_decoration with things like the perspective assistant
 
     AssistantEditorData toolData;
 
-    QPixmap m_iconDelete = KisIconUtils::loadIcon("dialog-cancel").pixmap(toolData.deleteIconSize, toolData.deleteIconSize);
+    QPixmap m_iconDelete = KisIconUtils::loadIcon("deletelayer").pixmap(toolData.deleteIconSize, toolData.deleteIconSize);
     QPixmap m_iconSnapOn = KisIconUtils::loadIcon("visible").pixmap(toolData.snapIconSize, toolData.snapIconSize);
     QPixmap m_iconSnapOff = KisIconUtils::loadIcon("novisible").pixmap(toolData.snapIconSize, toolData.snapIconSize);
     QPixmap m_iconMove = KisIconUtils::loadIcon("transform-move").pixmap(toolData.moveIconSize, toolData.moveIconSize);
+    QPixmap m_iconLockOn = KisIconUtils::loadIcon("layer-locked").pixmap(toolData.lockedIconSize, toolData.lockedIconSize);
+    QPixmap m_iconLockOff = KisIconUtils::loadIcon("layer-unlocked").pixmap(toolData.lockedIconSize, toolData.lockedIconSize);
+    QPixmap m_iconDragEditorWidget = KisIconUtils::loadIcon("gridbrush").pixmap(toolData.dragEditorWidgetIconSize, toolData.dragEditorWidgetIconSize);
+
+
 
     KisCanvas2 * m_canvas = 0;
 };
@@ -76,7 +74,9 @@ KisPaintingAssistantsDecoration::KisPaintingAssistantsDecoration(QPointer<KisVie
 {
     setAssistantVisible(true);
     setOutlineVisible(true);
+    setPriority(95);
     d->snapOnlyOneAssistant = true; //turn on by default.
+    d->snapEraser = false;
 }
 
 KisPaintingAssistantsDecoration::~KisPaintingAssistantsDecoration()
@@ -84,6 +84,15 @@ KisPaintingAssistantsDecoration::~KisPaintingAssistantsDecoration()
     delete d;
 }
 
+void KisPaintingAssistantsDecoration::slotUpdateDecorationVisibility()
+{
+    const bool shouldBeVisible = !assistants().isEmpty();
+
+    if (visible() != shouldBeVisible) {
+        setVisible(shouldBeVisible);
+    }
+}
+
 void KisPaintingAssistantsDecoration::addAssistant(KisPaintingAssistantSP assistant)
 {
     QList<KisPaintingAssistantSP> assistants = view()->document()->assistants();
@@ -119,64 +128,197 @@ void KisPaintingAssistantsDecoration::removeAll()
     emit assistantChanged();
 }
 
+void KisPaintingAssistantsDecoration::setAssistants(const QList<KisPaintingAssistantSP> &assistants)
+{
+    Q_FOREACH (KisPaintingAssistantSP assistant, assistants) {
+        assistant->setAssistantGlobalColorCache(view()->document()->assistantsGlobalColor());
+    }
+    view()->document()->setAssistants(assistants);
+    setVisible(!assistants.isEmpty());
+
+    emit assistantChanged();
+}
+
+void KisPaintingAssistantsDecoration::setAdjustedBrushPosition(const QPointF position)
+{
+    if (!assistants().empty()) {
+        Q_FOREACH (KisPaintingAssistantSP assistant, assistants()) {
+            assistant->setAdjustedBrushPosition(position);
+        }
+    }
+}
+
+
 QPointF KisPaintingAssistantsDecoration::adjustPosition(const QPointF& point, const QPointF& strokeBegin)
 {
 
     if (assistants().empty()) {
+        // No assistants, so no adjustment
         return point;
     }
 
+    if  (!d->snapEraser
+        && (d->m_canvas->resourceManager()->resource(KoCanvasResource::CurrentEffectiveCompositeOp).toString() == COMPOSITE_ERASE)) {
+        // No snapping if eraser snapping is disabled and brush is an eraser
+        return point;
+    }
+
+    KisImageSP image = d->m_canvas->image();
+    KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(image, point);
+
+    const KisCoordinatesConverter *converter = d->m_canvas->coordinatesConverter();
+    const qreal moveThresholdPt = 2.0 * converter->effectiveZoom() / qMax(image->xRes(), image->yRes());
+
+
+    // There is at least 1 assistant
     if (assistants().count() == 1) {
+        // Things are easy when there is only one assistant
         if(assistants().first()->isSnappingActive() == true){
-            QPointF newpoint = assistants().first()->adjustPosition(point, strokeBegin);
+            bool snapSingle = d->snapOnlyOneAssistant && d->aFirstStroke;
+            QPointF newpoint = assistants().first()->adjustPosition(point, strokeBegin, !snapSingle, moveThresholdPt);
             // check for NaN
             if (newpoint.x() != newpoint.x()) return point;
+            // Tell the assistant that its guidelines should
+            // follow the adjusted bush position.
+            assistants().first()->setFollowBrushPosition(true);
+
+            // we need setting d->aFirstStroke even for one assistant, because now
+            // single assistants can be made out of several inner ones (vide TwoPointPerspective).
+            //this is here to be compatible with the movement in the perspective tool.
+            qreal dx = point.x() - strokeBegin.x();
+            qreal dy = point.y() - strokeBegin.y();
+            if (dx * dx + dy * dy >= 4.0) {
+                // allow some movement before snapping
+                d->aFirstStroke=true;
+            }
             return newpoint;
+        } else {
+            // One assistant, but it is not active, so no adjustment
+            return point;
         }
     }
+
+    // There is more than one assistant.
+    // We have to chose one of these.
+    // This is done by checking which assistant gives the an adjusted position
+    // that is closest to the current position.
+
     QPointF best = point;
     double distance = DBL_MAX;
-    //the following tries to find the closest point to stroke-begin. It checks all assistants for the closest point//
-    if(!d->snapOnlyOneAssistant){
+    if (!d->snapOnlyOneAssistant) {
+        // In this mode the best assistant is constantly computed anew.
         Q_FOREACH (KisPaintingAssistantSP assistant, assistants()) {
-            if(assistant->isSnappingActive() == true){//this checks if the assistant in question has it's snapping boolean turned on//
-                QPointF pt = assistant->adjustPosition(point, strokeBegin);
+            if (assistant->isSnappingActive() == true){//this checks if the assistant in question has it's snapping boolean turned on//
+                QPointF pt = assistant->adjustPosition(point, strokeBegin, true, moveThresholdPt);
+                // check for NaN
                 if (pt.x() != pt.x()) continue;
                 double dist = qAbs(pt.x() - point.x()) + qAbs(pt.y() - point.y());
                 if (dist < distance) {
                     best = pt;
                     distance = dist;
                 }
+                assistant->setFollowBrushPosition(true);
             }
         }
     } else if (d->aFirstStroke==false) {
+        // In this mode we compute the best assistant during the initial
+        // movement and then keep using that assistant. (Called the first
+        // assistant).
+        bool foundAGoodAssistant = false;
         Q_FOREACH (KisPaintingAssistantSP assistant, assistants()) {
             if(assistant->isSnappingActive() == true){//this checks if the assistant in question has it's snapping boolean turned on//
-                QPointF pt = assistant->adjustPosition(point, strokeBegin);
+                QPointF pt = assistant->adjustPosition(point, strokeBegin, true, moveThresholdPt);
                 if (pt.x() != pt.x()) continue;
                 double dist = qAbs(pt.x() - point.x()) + qAbs(pt.y() - point.y());
                 if (dist < distance) {
                     best = pt;
                     distance = dist;
                     d->firstAssistant = assistant;
+                    foundAGoodAssistant = true;
                 }
+                assistant->setFollowBrushPosition(true);
             }
         }
-    } else if(d->firstAssistant) {
+        if (!foundAGoodAssistant) {
+            // reset the first assistant if none of them are available
+            // that helps in case all of the assistants are disabled
+            // BUG:448187
+            d->firstAssistant = 0;
+        }
+    } else if (d->firstAssistant) {
         //make sure there's a first assistant to begin with.//
-        best = d->firstAssistant->adjustPosition(point, strokeBegin);
+        QPointF newpoint = d->firstAssistant->adjustPosition(point, strokeBegin, false, moveThresholdPt);
+        // BUGFIX: 402535
+        // assistants might return (NaN,NaN), must always check for that
+        if (newpoint.x() == newpoint.x()) {
+            // not a NaN
+            best = newpoint;
+        }
     } else {
         d->aFirstStroke=false;
     }
+
     //this is here to be compatible with the movement in the perspective tool.
-    qreal dx = point.x() - strokeBegin.x(), dy = point.y() - strokeBegin.y();
+    qreal dx = point.x() - strokeBegin.x();
+    qreal dy = point.y() - strokeBegin.y();
     if (dx * dx + dy * dy >= 4.0) {
         // allow some movement before snapping
         d->aFirstStroke=true;
     }
+
     return best;
 }
 
+void KisPaintingAssistantsDecoration::adjustLine(QPointF &point, QPointF &strokeBegin)
+{
+    if (assistants().empty()) {
+        // No assistants, so no adjustment
+        return;
+    }
+
+    // TODO: figure it out
+    if  (!d->snapEraser
+        && (d->m_canvas->resourceManager()->resource(KoCanvasResource::CurrentEffectiveCompositeOp).toString() == COMPOSITE_ERASE)) {
+        // No snapping if eraser snapping is disabled and brush is an eraser
+        return;
+    }
+
+    QPointF originalPoint = point;
+    QPointF originalStrokeBegin = strokeBegin;
+
+    qreal minDistance = 10000.0;
+    bool minDistValid = false;
+    QPointF finalPoint = originalPoint;
+    QPointF finalStrokeBegin = originalStrokeBegin;
+    int id = 0;
+    KisPaintingAssistantSP bestAssistant;
+    Q_FOREACH (KisPaintingAssistantSP assistant, assistants()) {
+        if(assistant->isSnappingActive() == true){//this checks if the assistant in question has it's snapping boolean turned on//
+            //QPointF pt = assistant->adjustPosition(point, strokeBegin, true);
+            QPointF p1 = originalPoint;
+            QPointF p2 = originalStrokeBegin;
+            assistant->adjustLine(p1, p2);
+            if (p1.isNull() || p2.isNull()) {
+                // possibly lines cannot snap to this assistant, or this line cannot, at least
+                continue;
+            }
+            qreal distance = kisSquareDistance(p1, originalPoint) + kisSquareDistance(p2, originalStrokeBegin);
+            if (distance < minDistance || !minDistValid) {
+                finalPoint = p1;
+                finalStrokeBegin = p2;
+                minDistValid = true;
+                bestAssistant = assistant;
+            }
+        }
+        id ++;
+    }
+    if (bestAssistant) {
+        bestAssistant->setFollowBrushPosition(true);
+    }
+    point = finalPoint;
+    strokeBegin = finalStrokeBegin;
+}
+
 void KisPaintingAssistantsDecoration::endStroke()
 {
     d->aFirstStroke = false;
@@ -187,8 +329,8 @@ void KisPaintingAssistantsDecoration::endStroke()
 }
 
 void KisPaintingAssistantsDecoration::drawDecoration(QPainter& gc, const QRectF& updateRect, const KisCoordinatesConverter *converter,KisCanvas2* canvas)
-{    
-    if(assistants().length() == 0) {
+{
+    if(assistants().isEmpty()) {
         return; // no assistants to worry about, ok to exit
     }
 
@@ -199,15 +341,19 @@ void KisPaintingAssistantsDecoration::drawDecoration(QPainter& gc, const QRectF&
     }
 
     // the preview functionality for assistants. do not show while editing
-    if (d->m_isEditingAssistants) {
-        d->m_outlineVisible = false;
-    }
-    else {
-        d->m_outlineVisible = outlineVisibility();
-    }
+
+    KoToolProxy *proxy = view()->canvasBase()->toolProxy();
+    KIS_SAFE_ASSERT_RECOVER_RETURN(proxy);
+    KisToolProxy *kritaProxy = dynamic_cast<KisToolProxy*>(proxy);
+    KIS_SAFE_ASSERT_RECOVER_RETURN(kritaProxy);
+
+    const bool outlineVisible =
+        outlineVisibility() &&
+        !d->m_isEditingAssistants &&
+        kritaProxy->supportsPaintingAssistants();
 
     Q_FOREACH (KisPaintingAssistantSP assistant, assistants()) {
-        assistant->drawAssistant(gc, updateRect, converter, true, canvas, assistantVisibility(), d->m_outlineVisible);
+        assistant->drawAssistant(gc, updateRect, converter, true, canvas, assistantVisibility(), outlineVisible);
 
         if (isEditingAssistants()) {
             drawHandles(assistant, gc, converter);
@@ -290,10 +436,20 @@ QList<KisPaintingAssistantHandleSP> KisPaintingAssistantsDecoration::handles()
 
 QList<KisPaintingAssistantSP> KisPaintingAssistantsDecoration::assistants() const
 {
-    QList<KisPaintingAssistantSP> assistants = view()->document()->assistants();
+    QList<KisPaintingAssistantSP> assistants;
+    if (view()) {
+        if (view()->document()) {
+            assistants = view()->document()->assistants();
+        }
+    }
     return assistants;
 }
 
+bool KisPaintingAssistantsDecoration::hasPaintableAssistants() const
+{
+    return !assistants().isEmpty();
+}
+
 KisPaintingAssistantSP KisPaintingAssistantsDecoration::selectedAssistant()
 {
     return d->selectedAssistant;
@@ -326,6 +482,11 @@ void KisPaintingAssistantsDecoration::setOnlyOneAssistantSnap(bool assistant)
     d->snapOnlyOneAssistant = assistant;
 }
 
+void KisPaintingAssistantsDecoration::setEraserSnap(bool assistant)
+{
+    d->snapEraser = assistant;
+}
+
 bool KisPaintingAssistantsDecoration::assistantVisibility()
 {
     return d->assistantVisible;
@@ -348,6 +509,7 @@ void KisPaintingAssistantsDecoration::toggleAssistantVisible()
 
 void KisPaintingAssistantsDecoration::toggleOutlineVisible()
 {
+    // "outline" means assistant preview (line that depends on the mouse cursor)
     setOutlineVisible(!outlineVisibility());
 }
 
@@ -428,33 +590,32 @@ void KisPaintingAssistantsDecoration::drawEditorWidget(KisPaintingAssistantSP as
         return;
     }
 
+    AssistantEditorData toolData; // shared const data for positioning and sizing
+
     QTransform initialTransform = converter->documentToWidgetTransform();
 
-    // We are going to put all of the assistant actions below the bounds of the assistant
-    // so they are out of the way
-    // assistant->buttonPosition() gets the center X/Y position point
-    QPointF actionsPosition = initialTransform.map(assistant->buttonPosition());
-
-    AssistantEditorData toolData; // shared const data for positioning and sizing
+    QPointF actionsPosition = initialTransform.map(assistant->viewportConstrainedEditorPosition(converter, toolData.boundingSize));
 
     QPointF iconMovePosition(actionsPosition + toolData.moveIconPosition);
     QPointF iconSnapPosition(actionsPosition + toolData.snapIconPosition);
+    QPointF iconLockedPosition(actionsPosition + toolData.lockedIconPosition);
     QPointF iconDeletePosition(actionsPosition + toolData.deleteIconPosition);
+    QPointF iconDragEditorWidgetPosition(actionsPosition + toolData.dragEditorWidgetIconPosition);
 
     // Background container for helpers
-    QBrush backgroundColor = d->m_canvas->viewManager()->mainWindow()->palette().window();
+    QBrush backgroundColor = d->m_canvas->viewManager()->mainWindowAsQWidget()->palette().window();
     QPointF actionsBGRectangle(actionsPosition + QPointF(10, 10));
 
     gc.setRenderHint(QPainter::Antialiasing);
 
     QPainterPath bgPath;
-    bgPath.addRoundedRect(QRectF(actionsBGRectangle.x(), actionsBGRectangle.y(), 110, 40), 6, 6);
+    bgPath.addRoundedRect(QRectF(actionsBGRectangle.x(), actionsBGRectangle.y(), toolData.boundingSize.width(), toolData.boundingSize.height()), 6, 6);
     QPen stroke(QColor(60, 60, 60, 80), 2);
 
     // if the assistant is selected, make outline stroke fatter and use theme's highlight color
     // for better visual feedback
     if (selectedAssistant()) { // there might not be a selected assistant, so do not seg fault
-        if (assistant->buttonPosition() == selectedAssistant()->buttonPosition()) {
+        if (assistant->getEditorPosition() == selectedAssistant()->getEditorPosition()) {
             stroke.setWidth(4);
             stroke.setColor(qApp->palette().color(QPalette::Highlight));
         }
@@ -477,7 +638,18 @@ void KisPaintingAssistantsDecoration::drawEditorWidget(KisPaintingAssistantSP as
         gc.drawPixmap(iconSnapPosition, d->m_iconSnapOff);
     }
 
+    if (assistant->isLocked()) {
+        gc.drawPixmap(iconLockedPosition, d->m_iconLockOn);
+    } else {
+        qreal oldOpacity = gc.opacity();
+        gc.setOpacity(0.35);
+        gc.drawPixmap(iconLockedPosition, d->m_iconLockOff);
+        gc.setOpacity(oldOpacity);
+    }
+
+
     gc.drawPixmap(iconDeletePosition, d->m_iconDelete);
+    gc.drawPixmap(iconDragEditorWidgetPosition, d->m_iconDragEditorWidget);
 
 
 }
