@@ -1,20 +1,9 @@
 /*
- * Copyright (c) 2008 Cyrille Berger <cberger@cberger.net>
- * Copyright (c) 2010 Geoffry Song <goffrie@gmail.com>
- * Copyright (c) 2017 Scott Petrovic <scottpetrovic@gmail.com>
+ * SPDX-FileCopyrightText: 2008 Cyrille Berger <cberger@cberger.net>
+ * SPDX-FileCopyrightText: 2010 Geoffry Song <goffrie@gmail.com>
+ * SPDX-FileCopyrightText: 2017 Scott Petrovic <scottpetrovic@gmail.com>
  *
- *  This library is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU Lesser General Public License as published by
- *  the Free Software Foundation; version 2.1 of the License.
- *
- *  This library is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Lesser General Public License for more details.
- *
- *  You should have received a copy of the GNU Lesser General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: LGPL-2.0-or-later
  */
 
 #include "ConcentricEllipseAssistant.h"
@@ -22,6 +11,7 @@
 #include <klocalizedstring.h>
 #include "kis_debug.h"
 #include <QPainter>
+#include <QPainterPath>
 #include <QLinearGradient>
 #include <QTransform>
 #include <kis_canvas2.h>
@@ -31,7 +21,7 @@
 #include <math.h>
 
 ConcentricEllipseAssistant::ConcentricEllipseAssistant()
-        : KisPaintingAssistant("concentric ellipse", i18n("Concentric Ellipse assistant"))
+    : KisPaintingAssistant("concentric ellipse", i18n("Concentric Ellipse assistant"))
 {
 }
 
@@ -47,18 +37,16 @@ ConcentricEllipseAssistant::ConcentricEllipseAssistant(const ConcentricEllipseAs
 {
 }
 
-QPointF ConcentricEllipseAssistant::project(const QPointF& pt, const QPointF& strokeBegin) const
+QPointF ConcentricEllipseAssistant::project(const QPointF& pt, const QPointF& strokeBegin, const bool checkForInitialMovement, qreal moveThresholdPt) const
 {
     Q_ASSERT(isAssistantComplete());
     m_ellipse.set(*handles()[0], *handles()[1], *handles()[2]);
 
-    qreal
-            dx = pt.x() - strokeBegin.x(),
-            dy = pt.y() - strokeBegin.y();
-        if (dx * dx + dy * dy < 4.0) {
-            // allow some movement before snapping
-            return strokeBegin;
-        }
+    if (checkForInitialMovement && KisAlgebra2D::norm(pt - strokeBegin) < moveThresholdPt) {
+        // allow some movement before snapping
+        return strokeBegin;
+    }
+    //
     //calculate ratio
     QPointF initial = m_ellipse.project(strokeBegin);
     QPointF center = m_ellipse.boundingRect().center();
@@ -78,10 +66,14 @@ QPointF ConcentricEllipseAssistant::project(const QPointF& pt, const QPointF& st
     return m_extraEllipse.project(pt);
 }
 
-QPointF ConcentricEllipseAssistant::adjustPosition(const QPointF& pt, const QPointF& strokeBegin)
+QPointF ConcentricEllipseAssistant::adjustPosition(const QPointF& pt, const QPointF& strokeBegin, const bool /*snapToAny*/, qreal moveThresholdPt)
 {
-    return project(pt, strokeBegin);
+    return project(pt, strokeBegin, true, moveThresholdPt);
+}
 
+void ConcentricEllipseAssistant::adjustLine(QPointF &point, QPointF &strokeBegin)
+{
+    point = project(point, strokeBegin, false, 0.0);
 }
 
 void ConcentricEllipseAssistant::drawAssistant(QPainter& gc, const QRectF& updateRect, const KisCoordinatesConverter* converter, bool cached, KisCanvas2* canvas, bool assistantVisible, bool previewVisible)
@@ -100,11 +92,18 @@ void ConcentricEllipseAssistant::drawAssistant(QPainter& gc, const QRectF& updat
         dbgFile<<"canvas does not exist in the ellipse assistant, you may have passed arguments incorrectly:"<<canvas;
     }
 
-    QTransform initialTransform = converter->documentToWidgetTransform();
+
 
     if (isSnappingActive() && previewVisible == true){
 
         if (isAssistantComplete()){
+
+            QTransform initialTransform = converter->documentToWidgetTransform();
+
+            if (m_followBrushPosition && m_adjustedPositionValid) {
+                mousePos = initialTransform.map(m_adjustedBrushPosition);
+            }
+
             if (m_ellipse.set(*handles()[0], *handles()[1], *handles()[2])) {
                 QPointF initial = m_ellipse.project(initialTransform.inverted().map(mousePos));
                 QPointF center = m_ellipse.boundingRect().center();
@@ -172,7 +171,7 @@ QRect ConcentricEllipseAssistant::boundingRect() const
     }
 }
 
-QPointF ConcentricEllipseAssistant::buttonPosition() const
+QPointF ConcentricEllipseAssistant::getDefaultEditorPosition() const
 {
     return (*handles()[0] + *handles()[1]) * 0.5;
 }
@@ -182,6 +181,25 @@ bool ConcentricEllipseAssistant::isAssistantComplete() const
     return handles().size() >= 3;
 }
 
+void ConcentricEllipseAssistant::transform(const QTransform &transform)
+{
+    m_ellipse.set(*handles()[0], *handles()[1], *handles()[2]);
+
+    QPointF newAxes;
+    QTransform newTransform;
+
+    std::tie(newAxes, newTransform) = KisAlgebra2D::transformEllipse(QPointF(m_ellipse.semiMajor(), m_ellipse.semiMinor()), m_ellipse.getInverse() * transform);
+
+    const QPointF p1 = newTransform.map(QPointF(newAxes.x(), 0));
+    const QPointF p2 = newTransform.map(QPointF(-newAxes.x(), 0));
+    const QPointF p3 = newTransform.map(QPointF(0, newAxes.y()));
+
+    *handles()[0] = p1;
+    *handles()[1] = p2;
+    *handles()[2] = p3;
+
+    uncache();
+}
 
 ConcentricEllipseAssistantFactory::ConcentricEllipseAssistantFactory()
 {
