@@ -1,25 +1,13 @@
 /*
  *  kis_tool_line.cc - part of Krayon
  *
- *  Copyright (c) 2000 John Califf <jwcaliff@compuzone.net>
- *  Copyright (c) 2002 Patrick Julien <freak@codepimps.org>
- *  Copyright (c) 2003 Boudewijn Rempt <boud@valdyas.org>
- *  Copyright (c) 2009 Lukáš Tvrdý <lukast.dev@gmail.com>
- *  Copyright (c) 2007,2010 Cyrille Berger <cberger@cberger.net>
+ *  SPDX-FileCopyrightText: 2000 John Califf <jwcaliff@compuzone.net>
+ *  SPDX-FileCopyrightText: 2002 Patrick Julien <freak@codepimps.org>
+ *  SPDX-FileCopyrightText: 2003 Boudewijn Rempt <boud@valdyas.org>
+ *  SPDX-FileCopyrightText: 2009 Lukáš Tvrdý <lukast.dev@gmail.com>
+ *  SPDX-FileCopyrightText: 2007, 2010 Cyrille Berger <cberger@cberger.net>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "kis_tool_line.h"
@@ -38,15 +26,19 @@
 #include <kis_debug.h>
 #include <kis_cursor.h>
 #include <brushengine/kis_paintop_registry.h>
-#include "kis_figure_painting_tool_helper.h"
-#include "kis_canvas2.h"
+#include <kis_figure_painting_tool_helper.h>
+#include <kis_canvas2.h>
+#include <KisViewManager.h>
+#include <kis_action_registry.h>
+#include <kis_painting_information_builder.h>
 
-#include "kis_painting_information_builder.h"
 #include "kis_tool_line_helper.h"
 
+
 const KisCoordinatesConverter* getCoordinatesConverter(KoCanvasBase * canvas)
 {
     KisCanvas2 *kritaCanvas = dynamic_cast<KisCanvas2*>(canvas);
+    KIS_ASSERT(kritaCanvas);
     return kritaCanvas->coordinatesConverter();
 }
 
@@ -56,9 +48,11 @@ KisToolLine::KisToolLine(KoCanvasBase * canvas)
       m_showGuideline(true),
       m_strokeIsRunning(false),
       m_infoBuilder(new KisConverterPaintingInformationBuilder(getCoordinatesConverter(canvas))),
-      m_helper(new KisToolLineHelper(m_infoBuilder.data(), kundo2_i18n("Draw Line"))),
-      m_strokeUpdateCompressor(500, KisSignalCompressor::POSTPONE),
-      m_longStrokeUpdateCompressor(1000, KisSignalCompressor::FIRST_INACTIVE)
+      m_helper(new KisToolLineHelper(m_infoBuilder.data(),
+                                     canvas->resourceManager(),
+                                     kundo2_i18n("Draw Line"))),
+      m_strokeUpdateCompressor(200, KisSignalCompressor::POSTPONE),
+      m_longStrokeUpdateCompressor(750, KisSignalCompressor::FIRST_INACTIVE)
 {
     setObjectName("tool_line");
 
@@ -79,9 +73,9 @@ void KisToolLine::resetCursorStyle()
     overrideCursorIfNotEditable();
 }
 
-void KisToolLine::activate(ToolActivation activation, const QSet<KoShape*> &shapes)
+void KisToolLine::activate(const QSet<KoShape*> &shapes)
 {
-   KisToolPaint::activate(activation, shapes);
+   KisToolPaint::activate(shapes);
    configGroup =  KSharedConfig::openConfig()->group(toolId());
 }
 
@@ -104,16 +98,31 @@ QWidget* KisToolLine::createOptionWidget()
     m_chkShowGuideline = new QCheckBox(i18n("Show Guideline"));
     addOptionWidgetOption(m_chkShowGuideline);
 
+    m_chkSnapToAssistants = new QCheckBox(i18n("Snap to Assistants"));
+    addOptionWidgetOption(m_chkSnapToAssistants);
+
+    m_chkSnapEraser = new QCheckBox(i18n("Snap Eraser"));
+    addOptionWidgetOption(m_chkSnapEraser);
+
+
+
+
     // hook up connections for value changing
     connect(m_chkUseSensors, SIGNAL(clicked(bool)), this, SLOT(setUseSensors(bool)) );
     connect(m_chkShowPreview, SIGNAL(clicked(bool)), this, SLOT(setShowPreview(bool)) );
     connect(m_chkShowGuideline, SIGNAL(clicked(bool)), this, SLOT(setShowGuideline(bool)) );
+    connect(m_chkSnapToAssistants, SIGNAL(clicked(bool)), this, SLOT(setSnapToAssistants(bool)) );
 
 
     // read values in from configuration
     m_chkUseSensors->setChecked(configGroup.readEntry("useSensors", true));
     m_chkShowPreview->setChecked(configGroup.readEntry("showPreview", true));
     m_chkShowGuideline->setChecked(configGroup.readEntry("showGuideline", true));
+    m_chkSnapToAssistants->setChecked(configGroup.readEntry("snapToAssistants", false));
+    m_chkSnapEraser->setChecked(configGroup.readEntry("snapEraser", false));
+    if (!m_chkSnapToAssistants->isChecked()) {
+        m_chkSnapEraser->setEnabled(false);
+    }
 
     return widget;
 }
@@ -134,6 +143,17 @@ void KisToolLine::setShowPreview(bool value)
     configGroup.writeEntry("showPreview", value);
 }
 
+void KisToolLine::setSnapToAssistants(bool value)
+{
+    configGroup.writeEntry("snapToAssistants", value);
+    m_chkSnapEraser->setEnabled(value);
+}
+
+void KisToolLine::setSnapEraser(bool value)
+{
+    configGroup.writeEntry("snapEraser", value);
+}
+
 void KisToolLine::requestStrokeCancellation()
 {
     cancelStroke();
@@ -172,33 +192,46 @@ void KisToolLine::paint(QPainter& gc, const KoViewConverter &converter)
 void KisToolLine::beginPrimaryAction(KoPointerEvent *event)
 {
     NodePaintAbility nodeAbility = nodePaintAbility();
-    if (nodeAbility == NONE || !nodeEditable()) {
+    if (nodeAbility == UNPAINTABLE || !nodeEditable()) {
+        event->ignore();
+        return;
+    }
+
+    if (nodeAbility == MYPAINTBRUSH_UNPAINTABLE) {
+        KisCanvas2 * kiscanvas = static_cast<KisCanvas2*>(canvas());
+        QString message = i18n("The MyPaint Brush Engine is not available for this colorspace");
+        kiscanvas->viewManager()->showFloatingMessage(message, koIcon("object-locked"));
         event->ignore();
         return;
     }
 
     setMode(KisTool::PAINT_MODE);
 
+    const KisToolShape::ShapeAddInfo info =
+        shouldAddShape(currentNode());
+
     // Always show guideline on vector layers
     m_showGuideline = m_chkShowGuideline->isChecked() || nodeAbility != PAINT;
     updatePreviewTimer(m_showGuideline);
-    m_helper->setEnabled(nodeAbility == PAINT);
+    m_helper->setEnabled((nodeAbility == PAINT && !info.shouldAddShape) || info.shouldAddSelectionShape);
     m_helper->setUseSensors(m_chkUseSensors->isChecked());
     m_helper->start(event, canvas()->resourceManager());
 
     m_startPoint = convertToPixelCoordAndSnap(event);
     m_endPoint = m_startPoint;
     m_lastUpdatedPoint = m_startPoint;
+    m_originalStartPoint = m_startPoint;
 
     m_strokeIsRunning = true;
+
+    showSize();
 }
 
 void KisToolLine::updateStroke()
 {
     if (!m_strokeIsRunning) return;
 
-    m_helper->repaintLine(canvas()->resourceManager(),
-                          image(),
+    m_helper->repaintLine(image(),
                           currentNode(),
                           image().data());
 }
@@ -218,11 +251,14 @@ void KisToolLine::continuePrimaryAction(KoPointerEvent *event)
         m_helper->translatePoints(trans);
         m_startPoint += trans;
         m_endPoint += trans;
+        m_originalStartPoint += trans; // original start point is only original in terms of snapping to assistants
     } else if (event->modifiers() == Qt::ShiftModifier) {
         pos = straightLine(pos);
         m_helper->addPoint(event, pos);
     } else {
+        pos = snapToAssistants(pos);
         m_helper->addPoint(event, pos);
+        m_helper->movePointsTo(m_startPoint, pos);
     }
     m_endPoint = pos;
 
@@ -236,11 +272,22 @@ void KisToolLine::continuePrimaryAction(KoPointerEvent *event)
             m_longStrokeUpdateCompressor.stop();
             m_strokeUpdateCompressor.start();
             m_lastUpdatedPoint = pos;
-        } else if (updateDistance > 1) {
+        } else if (updateDistance > 1 &&  !m_strokeUpdateCompressor.isActive() && !m_longStrokeUpdateCompressor.isActive()) {
             m_longStrokeUpdateCompressor.start();
+            m_lastUpdatedPoint = pos;
         }
     }
 
+    if(event->modifiers() == Qt::AltModifier) {
+        KisCanvas2 *kisCanvas =dynamic_cast<KisCanvas2*>(canvas());
+        KIS_ASSERT(kisCanvas);
+        kisCanvas->viewManager()->showFloatingMessage(i18n("X: %1 px\nY: %2 px", QString::number(m_startPoint.x(), 'f',1)
+                                                           , QString::number(m_startPoint.y(), 'f',1))
+                                                           , QIcon(), 1000, KisFloatingMessage::High,  Qt::AlignLeft | Qt::TextWordWrap | Qt::AlignVCenter);
+    }
+    else {
+        showSize();
+    }
 
     updateGuideline();
     KisToolPaint::requestUpdateOutline(event->point, event);
@@ -254,17 +301,31 @@ void KisToolLine::endPrimaryAction(KoPointerEvent *event)
 
     updateGuideline();
     endStroke();
+
+    if (static_cast<KisCanvas2*>(canvas())->paintingAssistantsDecoration()) {
+        static_cast<KisCanvas2*>(canvas())->paintingAssistantsDecoration()->endStroke();
+    }
 }
 
+bool KisToolLine::primaryActionSupportsHiResEvents() const
+{
+    return true;
+}
+
+
 void KisToolLine::endStroke()
 {
     NodePaintAbility nodeAbility = nodePaintAbility();
 
-    if (!m_strokeIsRunning || m_startPoint == m_endPoint || nodeAbility == NONE) {
+    if (!m_strokeIsRunning || m_startPoint == m_endPoint || nodeAbility == UNPAINTABLE) {
+        m_helper->clearPoints();
         return;
     }
 
-    if (nodeAbility == PAINT) {
+    const KisToolShape::ShapeAddInfo info =
+        shouldAddShape(currentNode());
+
+    if ((nodeAbility == PAINT && !info.shouldAddShape) || info.shouldAddSelectionShape) {
         updateStroke();
         m_helper->end();
     }
@@ -281,7 +342,7 @@ void KisToolLine::endStroke()
         KoShapeStrokeSP border(new KoShapeStroke(currentStrokeWidth(), currentFgColor().toQColor()));
         path->setStroke(border);
 
-        KUndo2Command * cmd = canvas()->shapeController()->addShape(path, 0);
+        KUndo2Command * cmd = canvas()->shapeController()->addShape(path, nullptr);
         canvas()->addCommand(cmd);
     }
 
@@ -303,7 +364,6 @@ void KisToolLine::cancelStroke()
         m_helper->cancel();
     }
 
-
     m_strokeIsRunning = false;
     m_endPoint = m_startPoint;
 }
@@ -331,6 +391,31 @@ QPointF KisToolLine::straightLine(QPointF point)
     return result;
 }
 
+QPointF KisToolLine::snapToAssistants(QPointF point)
+{
+    if (m_chkSnapToAssistants->isChecked() && static_cast<KisCanvas2*>(canvas())->paintingAssistantsDecoration()) {
+        KisCanvas2* c = static_cast<KisCanvas2*>(canvas());
+        c->paintingAssistantsDecoration()->setOnlyOneAssistantSnap(true);
+        c->paintingAssistantsDecoration()->setEraserSnap(m_chkSnapEraser->isChecked());
+        QPointF startPoint = m_originalStartPoint;
+
+        // startPoint etc. are in image coordinates system (pixels)
+        // but assistants work in document coordinates system ("points")
+        QPointF startPointInDoc = getCoordinatesConverter(canvas())->imageToDocument(startPoint);
+        QPointF pointInDoc = getCoordinatesConverter(canvas())->imageToDocument(point);
+
+        c->paintingAssistantsDecoration()->adjustLine(pointInDoc, startPointInDoc);
+        c->paintingAssistantsDecoration()->setAdjustedBrushPosition(pointInDoc);
+
+        startPoint = getCoordinatesConverter(canvas())->documentToImage(startPointInDoc);
+        point = getCoordinatesConverter(canvas())->documentToImage(pointInDoc);
+
+        m_startPoint = startPoint;
+        return point;
+    }
+    return point;
+}
+
 
 void KisToolLine::updateGuideline()
 {
@@ -341,6 +426,13 @@ void KisToolLine::updateGuideline()
 }
 
 
+void KisToolLine::showSize()
+{
+    KisCanvas2 *kisCanvas =dynamic_cast<KisCanvas2*>(canvas());
+    KIS_ASSERT(kisCanvas);
+    kisCanvas->viewManager()->showFloatingMessage(i18n("Length: %1 px", QString::number(QLineF(m_startPoint,m_endPoint).length(), 'f',1))
+                                                        , QIcon(), 1000, KisFloatingMessage::High,  Qt::AlignLeft | Qt::TextWordWrap | Qt::AlignVCenter);
+}
 void KisToolLine::paintLine(QPainter& gc, const QRect&)
 {
     QPointF viewStartPos = pixelToView(m_startPoint);
@@ -359,4 +451,7 @@ QString KisToolLine::quickHelp() const
     return i18n("Alt+Drag will move the origin of the currently displayed line around, Shift+Drag will force you to draw straight lines");
 }
 
-
+bool KisToolLine::supportsPaintingAssistants() const
+{
+    return true;
+}
