@@ -1,20 +1,8 @@
 /*
  *
- *  Copyright (c) 2007 Sven Langkamp <sven.langkamp@gmail.com>
+ *  SPDX-FileCopyrightText: 2007 Sven Langkamp <sven.langkamp@gmail.com>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #ifndef KIS_TOOL_MEASURE_H_
@@ -42,20 +30,22 @@ class KisToolMeasureOptionsWidget : public QWidget
     Q_OBJECT
 
 public:
-    KisToolMeasureOptionsWidget(QWidget* parent, double resolution);
+    KisToolMeasureOptionsWidget(QWidget* parent, KisImageWSP image);
 
 public Q_SLOTS:
     void slotSetDistance(double distance);
     void slotSetAngle(double angle);
     void slotUnitChanged(int index);
+    void slotResolutionChanged(double xRes, double yRes);
 
 private:
     void updateDistance();
 
+    QLabel* m_angleLabel {nullptr};
+    double m_distance {0.0};
+public:
+    QLabel* m_distanceLabel {nullptr};
     double m_resolution;
-    QLabel* m_distanceLabel;
-    QLabel* m_angleLabel;
-    double m_distance;
     KoUnit m_unit;
 };
 
@@ -71,6 +61,9 @@ public:
     void beginPrimaryAction(KoPointerEvent *event) override;
     void continuePrimaryAction(KoPointerEvent *event) override;
     void endPrimaryAction(KoPointerEvent *event) override;
+    void showDistanceAngleOnCanvas();
+
+    QPointF lockedAngle(QPointF pos);
 
     void paint(QPainter& gc, const KoViewConverter &converter) override;
 
@@ -93,7 +86,7 @@ private:
     }
 
 private:
-    KisToolMeasureOptionsWidget *m_optionsWidget;
+    KisToolMeasureOptionsWidget *m_optionsWidget {nullptr};
 
     QPointF m_startPos {QPointF(0, 0)};
     QPointF m_endPos {QPointF(0, 0)};
@@ -107,7 +100,7 @@ public:
 
     KisToolMeasureFactory()
             : KoToolFactoryBase("KritaShape/KisToolMeasure") {
-        setSection(TOOL_TYPE_VIEW);
+        setSection(ToolBoxSection::View);
         setToolTip(i18n("Measure Tool"));
         setIconName(koIconNameCStr("krita_tool_measure"));
         setPriority(1);
