@@ -1,19 +1,8 @@
 /*
- *  Copyright (c) 2014 Dmitry Kazakov <dimula73@gmail.com>
+ *  SPDX-FileCopyrightText: 2014 Dmitry Kazakov <dimula73@gmail.com>
+ *  SPDX-FileCopyrightText: 2022 Carsten Hartenfels <carsten.hartenfels@pm.me>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "kis_perspective_transform_strategy.h"
@@ -43,6 +32,18 @@ enum StrokeFunction {
     NONE
 };
 
+enum HandleIndexes {
+    HANDLE_TOP_LEFT = 0,
+    HANDLE_TOP_RIGHT,
+    HANDLE_BOTTOM_LEFT,
+    HANDLE_BOTTOM_RIGHT,
+    HANDLE_MIDDLE_TOP,
+    HANDLE_MIDDLE_BOTTOM,
+    HANDLE_MIDDLE_LEFT,
+    HANDLE_MIDDLE_RIGHT,
+    HANDLE_COUNT,
+};
+
 struct KisPerspectiveTransformStrategy::Private
 {
     Private(KisPerspectiveTransformStrategy *_q,
@@ -80,11 +81,11 @@ struct KisPerspectiveTransformStrategy::Private
 
     /// custom members ///
 
-    StrokeFunction function;
+    StrokeFunction function {NONE};
 
     struct HandlePoints {
-        bool xVanishingExists;
-        bool yVanishingExists;
+        bool xVanishingExists {false};
+        bool yVanishingExists {false};
 
         QPointF xVanishing;
         QPointF yVanishing;
@@ -93,15 +94,15 @@ struct KisPerspectiveTransformStrategy::Private
 
     QTransform transform;
 
-    QVector<QPointF> srcCornerPoints;
-    QVector<QPointF> dstCornerPoints;
-    int currentDraggingCornerPoint;
+    QVector<QPointF> srcHandlePoints;
+    QVector<QPointF> dstHandlePoints;
+    int currentDraggingHandlePoint {0};
 
-    bool imageTooBig;
+    bool imageTooBig {false};
 
     QPointF clickPos;
     ToolTransformArgs clickArgs;
-    bool isTransforming;
+    bool isTransforming {false};
 
     QCursor getScaleCursor(const QPointF &handlePt);
     QCursor getShearCursor(const QPointF &start, const QPointF &end);
@@ -127,15 +128,19 @@ KisPerspectiveTransformStrategy::~KisPerspectiveTransformStrategy()
 
 void KisPerspectiveTransformStrategy::Private::recalculateTransformedHandles()
 {
-    srcCornerPoints.clear();
-    srcCornerPoints << transaction.originalTopLeft();
-    srcCornerPoints << transaction.originalTopRight();
-    srcCornerPoints << transaction.originalBottomLeft();
-    srcCornerPoints << transaction.originalBottomRight();
+    srcHandlePoints.resize(HANDLE_COUNT);
+    srcHandlePoints[HANDLE_TOP_LEFT] = transaction.originalTopLeft();
+    srcHandlePoints[HANDLE_TOP_RIGHT] = transaction.originalTopRight();
+    srcHandlePoints[HANDLE_BOTTOM_LEFT] = transaction.originalBottomLeft();
+    srcHandlePoints[HANDLE_BOTTOM_RIGHT] = transaction.originalBottomRight();
+    srcHandlePoints[HANDLE_MIDDLE_TOP] = transaction.originalMiddleTop();
+    srcHandlePoints[HANDLE_MIDDLE_BOTTOM] = transaction.originalMiddleBottom();
+    srcHandlePoints[HANDLE_MIDDLE_LEFT] = transaction.originalMiddleLeft();
+    srcHandlePoints[HANDLE_MIDDLE_RIGHT] = transaction.originalMiddleRight();
 
-    dstCornerPoints.clear();
-    Q_FOREACH (const QPointF &pt, srcCornerPoints) {
-        dstCornerPoints << transform.map(pt);
+    dstHandlePoints.clear();
+    Q_FOREACH (const QPointF &pt, srcHandlePoints) {
+        dstHandlePoints << transform.map(pt);
     }
 
     QMatrix4x4 realMatrix(transform);
@@ -175,12 +180,12 @@ void KisPerspectiveTransformStrategy::setTransformFunction(const QPointF &mouseP
                                   handleRadius, DRAG_Y_VANISHING_POINT);
     }
 
-    m_d->currentDraggingCornerPoint = -1;
-    for (int i = 0; i < m_d->dstCornerPoints.size(); i++) {
-        if (handleChooser.addFunction(m_d->dstCornerPoints[i],
+    m_d->currentDraggingHandlePoint = -1;
+    for (int i = 0; i < m_d->dstHandlePoints.size(); i++) {
+        if (handleChooser.addFunction(m_d->dstHandlePoints[i],
                                       handleRadius, DRAG_HANDLE)) {
 
-            m_d->currentDraggingCornerPoint = i;
+            m_d->currentDraggingHandlePoint = i;
         }
     }
 
@@ -241,6 +246,10 @@ void KisPerspectiveTransformStrategy::paint(QPainter &gc)
     addHandleRectFunc(m_d->transaction.originalTopRight());
     addHandleRectFunc(m_d->transaction.originalBottomLeft());
     addHandleRectFunc(m_d->transaction.originalBottomRight());
+    addHandleRectFunc(m_d->transaction.originalMiddleTop());
+    addHandleRectFunc(m_d->transaction.originalMiddleBottom());
+    addHandleRectFunc(m_d->transaction.originalMiddleLeft());
+    addHandleRectFunc(m_d->transaction.originalMiddleRight());
 
     gc.save();
 
@@ -330,11 +339,11 @@ Eigen::Matrix3f getTransitionMatrix(const QVector<QPointF> &sp)
     Eigen::Matrix3f A;
     Eigen::Vector3f v3;
 
-    A << sp[0].x() , sp[1].x() , sp[2].x()
-        ,sp[0].y() , sp[1].y() , sp[2].y()
-        ,      1   ,       1   ,       1;
+    A << sp[HANDLE_TOP_LEFT].x() , sp[HANDLE_TOP_RIGHT].x() , sp[HANDLE_BOTTOM_LEFT].x()
+        ,sp[HANDLE_TOP_LEFT].y() , sp[HANDLE_TOP_RIGHT].y() , sp[HANDLE_BOTTOM_LEFT].y()
+        ,                    1 ,                        1   ,                        1;
 
-    v3 << sp[3].x() , sp[3].y() , 1;
+    v3 << sp[HANDLE_BOTTOM_RIGHT].x() , sp[HANDLE_BOTTOM_RIGHT].y() , 1;
 
     Eigen::Vector3f coeffs = A.colPivHouseholderQr().solve(v3);
 
@@ -409,40 +418,41 @@ void KisPerspectiveTransformStrategy::Private::transformIntoArgs(const Eigen::Ma
 
     m = T.inverse() * m;
 
-    // TODO: implement matrix decomposition as described here
+    /**
+     * We disabled decomposed transformation due to bug
+     * https://bugs.kde.org/show_bug.cgi?id=447255
+     *
+     * In some cases decomposed preliminary transformation
+     * shrinks the image into a very small size, which is later
+     * inflated by the perspective transform. It creates a really
+     * bad and blurry result.
+     *
+     * Even though the usage of preliminary rotation makes the bug
+     * much less obvious, but the image is still really blurred.
+     */
+
+#if 0
+    // Decomposition according to:
     // https://www.w3.org/TR/css-transforms-1/#decomposing-a-3d-matrix
+    KisAlgebra2D::DecomposedMatrix dm(toQTransform(m));
 
-    // For now use an extremely hackish approximation
-    if (m(0,1) != 0.0 && m(0,0) != 0.0 && m(2,2) != 0.0) {
+    currentArgs.setScaleX(dm.scaleX);
+    currentArgs.setScaleY(dm.scaleY);
 
-        const qreal factor = (m(1,1) / m(0,1) - m(1,0) / m(0,0));
+    currentArgs.setShearX(dm.shearXY);
+    currentArgs.setShearY(0.0);
 
-        qreal scaleX = m(0,0) / m(2,2);
-        qreal scaleY = m(0,1) / m(2,2) * factor;
+    currentArgs.setAZ(kisDegreesToRadians(dm.angle));
 
-        Eigen::Matrix3f SC = fromScale(scaleX, scaleY);
-
-        qreal shearX = 1.0 / factor;
-        qreal shearY = m(1,0) / m(0,0);
-
-        Eigen::Matrix3f S = fromShear(shearX, shearY);
-
-        currentArgs.setScaleX(scaleX);
-        currentArgs.setScaleY(scaleY);
-
-        currentArgs.setShearX(shearX);
-        currentArgs.setShearY(shearY);
-
-        m = m * SC.inverse();
-        m = m * S.inverse();
-        m /= m(2,2);
-    } else {
-        currentArgs.setScaleX(1.0);
-        currentArgs.setScaleY(1.0);
-
-        currentArgs.setShearX(0.0);
-        currentArgs.setShearY(0.0);
-    }
+    QTransform pre = dm.scaleTransform() * dm.shearTransform() * dm.rotateTransform();
+    m = m * fromQTransform(pre.inverted());
+#else
+    currentArgs.setScaleX(1.0);
+    currentArgs.setScaleY(1.0);
+    currentArgs.setShearX(0.0);
+    currentArgs.setShearY(0.0);
+    currentArgs.setAZ(0.0);
+#endif
 
     currentArgs.setTransformedCenter(QPointF(tX, tY));
     currentArgs.setFlattenedPerspectiveTransform(toQTransform(m));
@@ -462,9 +472,9 @@ QPointF toQPointF(const QVector4D &v) {
     return v.toVector2DAffine().toPointF();
 }
 
-void KisPerspectiveTransformStrategy::continuePrimaryAction(const QPointF &mousePos, bool shiftModifierActve, bool altModifierActive)
+void KisPerspectiveTransformStrategy::continuePrimaryAction(const QPointF &mousePos, bool shiftModifierActive, bool altModifierActive)
 {
-    Q_UNUSED(shiftModifierActve);
+    Q_UNUSED(shiftModifierActive);
     Q_UNUSED(altModifierActive);
 
     m_d->isTransforming = true;
@@ -479,11 +489,36 @@ void KisPerspectiveTransformStrategy::continuePrimaryAction(const QPointF &mouse
         break;
     }
     case DRAG_HANDLE: {
-        KIS_ASSERT_RECOVER_RETURN(m_d->currentDraggingCornerPoint >=0);
-        m_d->dstCornerPoints[m_d->currentDraggingCornerPoint] = mousePos;
+        KIS_ASSERT_RECOVER_RETURN(m_d->currentDraggingHandlePoint >= 0);
+        KIS_ASSERT_RECOVER_RETURN(m_d->currentDraggingHandlePoint < HANDLE_COUNT);
+        if (m_d->currentDraggingHandlePoint < HANDLE_MIDDLE_TOP) {
+            // Corner point, transform directly.
+            m_d->dstHandlePoints[m_d->currentDraggingHandlePoint] = mousePos;
+        } else {
+            // Middle point, move adjacent corners.
+            QPointF delta = mousePos - m_d->dstHandlePoints[m_d->currentDraggingHandlePoint];
+            switch(m_d->currentDraggingHandlePoint) {
+            case HANDLE_MIDDLE_TOP:
+                m_d->dstHandlePoints[HANDLE_TOP_LEFT] += delta;
+                m_d->dstHandlePoints[HANDLE_TOP_RIGHT] += delta;
+                break;
+            case HANDLE_MIDDLE_BOTTOM:
+                m_d->dstHandlePoints[HANDLE_BOTTOM_LEFT] += delta;
+                m_d->dstHandlePoints[HANDLE_BOTTOM_RIGHT] += delta;
+                break;
+            case HANDLE_MIDDLE_LEFT:
+                m_d->dstHandlePoints[HANDLE_TOP_LEFT] += delta;
+                m_d->dstHandlePoints[HANDLE_BOTTOM_LEFT] += delta;
+                break;
+            case HANDLE_MIDDLE_RIGHT:
+                m_d->dstHandlePoints[HANDLE_TOP_RIGHT] += delta;
+                m_d->dstHandlePoints[HANDLE_BOTTOM_RIGHT] += delta;
+                break;
+            }
+        }
 
-        Eigen::Matrix3f A = getTransitionMatrix(m_d->srcCornerPoints);
-        Eigen::Matrix3f B = getTransitionMatrix(m_d->dstCornerPoints);
+        Eigen::Matrix3f A = getTransitionMatrix(m_d->srcHandlePoints);
+        Eigen::Matrix3f B = getTransitionMatrix(m_d->dstHandlePoints);
         Eigen::Matrix3f result = B * A.inverse();
 
         m_d->transformIntoArgs(result);
@@ -692,4 +727,5 @@ void KisPerspectiveTransformStrategy::Private::recalculateTransformations()
     recalculateTransformedHandles();
 
     emit q->requestShowImageTooBig(imageTooBig);
+    emit q->requestImageRecalculation();
 }
